{
This unit is designed for Free Pascal Compiler!!

This unit contains types and constants
(taken from Sys,Objects,Storage,Images) required to create external module
libraries. Use this file only for low-level modules where you don't need to
use units mentioned above (they would increase size of your library n-times),
but you need some type/consts declared there.


All modules which uses some other libraries should be placed in external modules,
not in main project, so engine core requires no libraries to run.
}


unit ExtTypesFPC;



{$PACKENUM 4}
{Delphi Compatibility must be set ON}


interface


type


{*************************************************************************}
{                           From Sys.pas                                  }
{*************************************************************************}


  {Some integer types}
  DWord = Cardinal;
  ULong = Cardinal;
  Small = SmallInt;
  Integer = LongInt;
  Short = ShortInt;
  Long = LongInt;

  {Some pointer types}
  PLong = ^Long;
  PShort = ^Short;
  PSmall = ^Small;
  PDWord = ^DWord;
  PULong = ^ULong;


  {Floating point types.
   To change precision use Double}
  Float = Single;
  //Float = Double;

  PFloat = ^Float;


  {String and character types}
  // string = AnsiString
  WString = WideString;
  WChar = WideChar;

  PString = ^string;
  PWString = ^WString;
  PWChar = ^WChar;


  {some other types}
  THandle = ULong;


  {Version info record}
  TVersionInfo = packed record
    case Integer of
      0: (
        Major: Byte;
        Minor: Byte;
        Build: Word; );
      1: (
        Version: ULong);
  end;


  {Platform info . It is used e.g. by external modules to
  define for which platform it is designed}
  TPlatform = (
    pfUnknown,      // uknown or unimportant
    pfWin32,        // 32bit Windows (for modules it means  "Win32 only")
    pfWin64,        // 64bit Windows
    pfLinux,        // Linux  (for modules it means  "Linux only")
    pfIndependent   // e.g. for modules which are cross-platform
    );


  {Message structure used for comunication between objects and other
  purposes.}
  TMsgRec = packed record
    MsgID: ULong;            // Message ID constant. According to this ID
                             // procedures which performs this message
                             // access other memers of structure.
                             // E.g. Object receibes SM_FreeSender so
                             // Sender can not be nil !
    Sender: TObject;         // Identifies object that send this message
    case Long of
      0: (Data: Pointer);    // Custom pointer data
      1: (ULParam: ULong);   // Custom unsigned Long value
      2: (LParam: Long);     // Custom Long value
      3: (Words: array[0..1] of Word);  // Custom word values
      4: (Bytes: array[0..3] of Byte);  // Custom byte values
      5: (Chars: array[0..3] of Char);  // Custom chars
  end;


{*************************************************************************}
{                       From Objects.pas                                  }
{*************************************************************************}


const

  {Names of functions which are required in external module libraries.}
  sModuleCount = 'ModuleCount';
  sModuleInfo = 'ModuleInfo';
  sModuleExtInfo = 'ModuleExtInfo';

  {Mask which is used when searching for modules in External Module Libraries}
  sModuleFileMask = 'mod*.*';


type

  {This code is used for identification of modules e.g. 'IMG-DCIF','MUS-MIDI'.
  First three characters determine type of Module.
  It is case-sensitive.
  IMG is ImageModule, MUS is MusicModule, ...
  These chars are defined near declaration of adequate module.
  So IMG is in Images.pas}
  TIDCode = array[0..7] of Char;


  {This enum describes style of module implementation.}
  TModuleStyle = (
    msInternal,       // Module is internal. It is compiled with entire engine and
                      // module's source is in subdir of dir containing source
                      // of module's abstract ancestor class. For example:
                      // TJpgModule is in JpgModule.pas which is in
                      // Engine\Data\Images\Jpg directory
    msLowLevel,       // Module is wrapper to low-level functions in external library
    msAbsWrapper,     // Module is abstract unassigned wrapper to low-level module.
                      // No instance of this module is ever created, but when
                      // creating instance of wrapper to specific low-level module
                      // ClassRef from this module is used (based on IDCode
                      // agreement)
    msInterface,      // Module is accessed trough  interface from ext. library
    msClassIns,       // Module library returns created instance of module's class
    msClassRef        // Module library returns reference to module's class.
                      // Instance of this class must be created then
    );


  {Record with information about module. Mainly for communicating with
  modules in External Module libraries.}
  TModuleInfo = packed record
    Name: array[0..31] of Char;          // module name e.g. "Zip Module".
                                         // Names are unique !!
                                         // Remember this when creating new one.
    Description: array[0..95] of Char;   // short module desc e.g.
                                         // "Provides access to Zip (*.zip)
                                         // archives"
    LibName: packed array[0..31] of Char;// name of module library.
                                         // Must begin with "Mod"
                                         // e.g. "ModZipArchive.dll"
    Creator: array[0..63] of Char;       // author name, group, ...
                                         // e.g. "Hasan Ahmed,
                                         // Taliban Explosive Computing"
    Language: array[0..15] of Char;      // language which this module
                                         // was written in e.g. "Delphi 7"
    IDCode: TIDCode;                     // see info about TIDCode
    Style: TModuleStyle;                 // indicates style of module
                                         // implementation
    Version: TVersionInfo;               // current module version
    Platform: TPlatform;                 // platform on which module is
                                         // able to run
    RequiresLib: boolean;                // if true, module requires
                                         // another library (BPL/DLL/SO) to run
  end;

  PModuleInfo = ^TModuleInfo;


  {Function types which are called from module libraries. They are declared as
  cdecl so they must be marked as cdecl in ext. libaries too.
  It is because of some compilers (mainly C++) exports functions
  under different name than declared (use __cdecl and extern "c" to
  avoid this).
  Each module library - even high level - must contain these three }

  {Returns total count of module implementors in library.}
  TModuleCountProc = procedure(var Count: Long); cdecl;

  {Returns base information of given module.}
  TModuleInfoProc = procedure(Index: Long; var Info: TModuleInfo); cdecl;

  {Returns extended information on given module.
  Differs for each Module descendant classes. Pointer
  to record with fields and functions for low-level functions or
  Pointer to function which creates instance of high-level module.
  Functions returned by ExtInfo (in structure or directly) all uses
  register calling (__fastcall), because this is Delphi's native
  and here we don't need exact names no more.
  !!! Don't call FreeMem on returned Pointer}
  TModuleExtInfoProc = procedure(Index: Long; var ExtInfo: Pointer); cdecl;


{*************************************************************************}
{                       From Storage.pas                                  }
{*************************************************************************}


const

  {ID Code for descendants of TCompressModule is 'CMP' so modules can
  have IDcodes like: CMP-RLE, CMP-BZ2, ...}
  CompressModuleID = 'CMP';


type

  {Remember that these functions must be given allocated
  OutData when called. They can not call GetMem(OutData) because
  later it will cause Error when you try to Free this memory
  (it belongs to another process). Use simple GetMem(OutData,1) before call.
  If you are using TExtCompressModule it will be done automaticaly for you.}
 
  {Compress buffer function type}
  TCompressBufferProc = function(InData: Pointer; InSize: Long; var OutData:
    Pointer; var OutSize: Long): boolean; cdecl;


  {Decompress buffer function type}
  TDecompressBufferProc = function(InData: Pointer; InSize: Long; var OutData:
    Pointer; var OutSize: Long): boolean; cdecl;


  {This record is returned from Module Library (dll/bpl/so) to TExtCompressModule
  and this class then assigns returned functions to self functions .}
  TExtCompressInfo = packed record
    CanCompress: boolean;
    CanDecompress: boolean;
    CompressBuffer: TCompressBufferProc;
    DecompressBuffer: TDecompressBufferProc;
  end;

  PExtCompressInfo = ^TExtCompressInfo;


{*************************************************************************}
{                       From Images.pas                                   }
{*************************************************************************}


const

   {ID Code for descendants of TImageModule is 'IMG' so modules can
  have IDcodes like: IMG-BMP, IMG-MNG, ...}
  sImageModuleID = 'IMG';


implementation

end.

