{*************************************************************************}
{*   BmpModule                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images\Bmp\BmpModule.pas                                    *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 22th March 2003  (22.03.2003)                              }
{*************************************************************************}


{This unit contains module for loading/saving Bitmap files.}


unit BmpModule;


interface


uses
  Sys,Images,Objects,Maths,SysUtils,Storage;


type


  {Image encoder for loading/saving bitmaps in BMP/DIB format.
  Supports: 1,4,8,15,16,24,32 bit bitmaps loading/saving (OS2 format loading)
  Compression: RLE4, RLE8 loading/saving }
  TBmpModule = class(TImageModule)
  protected

    FUseRLE: boolean;

    function GetUseRLE: boolean;

    procedure SetUseRLE(Value: boolean);

    {Here convert from image formats not supported by encoder }
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      override;


  public

    constructor Create; override;
    destructor Destroy; override;

    {This function converts ImageData structure to bitmap.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This func takes valid image data and converts it to bitmap
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;


    {This function converts bitmap data to ImageData structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

     {This func converts bitmap in Input stream read from bitmap file to  Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;


    {This function tests if data are valid bitmap image.}
    function TestFormat(Data: Pointer; Size: Long): boolean; override;

    {This function tests if stream contains valid bitmap.}
    function TestFormat(Input: TStream): boolean; override;


  published

     {Indicates that module uses RLE compression when saving bitmpas.}
    property UseRLE: boolean read GetUseRLE write SetUSeRLE;

  end;


const

  {BMP file identifier}
  BMPMagic:array[0..1]of Char = 'BM';

  { constants for the TBitmapInfoHeader.Compression field }
  BI_RGB = 0;
  BI_RLE8 = 1;
  BI_RLE4 = 2;
  BI_BITFIELDS = 3;


type

  {File Header for Windows/OS2 bitmap file.}
  TBitMapFileHeader = packed record
    ID: Word; // Is always 19778 : 'BM'
    Size: ULong; // Filesize
    Reserved1: Word;
    Reserved2: Word;
    Offset: ULong; // Offset of image data
  end;


  {Info Header for Windows bitmap file.}
  TBitMapInfoHeader = packed record
    Size: ULong;
    Width: Long;
    Height: Long;
    Planes: Word;
    BitCount: Word;
    Compression: ULong;
    SizeImage: ULong;
    XPelsPerMeter: Long;
    YPelsPerMeter: Long;
    ClrUsed: ULong;
    ClrImportant: ULong;
  end;


  {Info Header for OS2 bitmaps}
  TBitmapCoreHeader = packed record
    Size: ULong;
    Width: Word;
    Height: Word;
    Planes: Word;
    BitCount: Word;
  end;

  {Used with BitmapInfo.Compression = BI_BITFIELDS}
  TLocalPixelFormat = packed record
    RBitMask, GBitMask, BBitMask: ULong;
  end;


const

  sBmpName = 'Bmp Module';
  sBmpDesc =
    'Provides Windows (or OS2) Bitmap (*.bmp) image format encoding/decoding';
  sBmpCode: TIDCode = 'IMG-BMP';


var

  BmpModuleInfo:TModuleInfo;


implementation


{*************************************************************************}
{                         TBmpModule Methods                              }
{*************************************************************************}


constructor TBmpModule.Create;
begin
  inherited Create;
  FInfo := BmpModuleInfo;

  FUseRLE := true;
end;

destructor TBmpModule.Destroy;
begin

  inherited Destroy;
end;

function TBmpModule.GetUseRLE: boolean;
begin
  Result := FUseRLE;
end;

procedure TBmpModule.SetUseRLE(Value: boolean);
begin
  FUseRLE := Value;
end;

procedure TBmpModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone do
  begin
   // BMP stores images as flipped with dword aligned scanlines
   // 16bit images can be saved -> it supports custom pixel formats
   // in all 8bit images must be palette present
    case ImageFormat of
     ifIntensity8: Convert8To8(Clone,ifIndex8);
     Low(if8bitARGBPF) .. High(if8bitARGBPF): Convert8To24(Clone);
    end;
    FlipYImage(Clone,0);
  end;
end;

function TBmpModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TBmpModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  BF: TBitmapFileHeader;
  BI: TBitmapInfoHeader;
  StartPos: Int64;
  LocalPF: TLocalPixelFormat;
  FSize, Off: Long;
  WidthBytes: Long;
  Data: Pointer;
  DSize: Long;
  Clone: TImageData;

  procedure SaveRGB;
  begin
    OutPut.WriteBuffer(Data^, DSize);
  end;

  procedure SaveRLE8;
  var
    Size: Long;
    B1, B2: Byte;
    PB1, PB2: Long;
    Src, Buff: PByte;
    X, Y: Long;

    function AllocByte: PByte;
    begin
      if Size mod 4096 = 0 then
        ReAllocMem(Buff, Size + 4095);
      Result := Pointer(Long(Buff) + Size);
      Size := Size + 1;
    end;

  begin
    // not necessary
    GetMem(Buff, 1024);
    try
      Size := 0;
      for Y := 0 to Clone.Levels[0].Height - 1 do
      begin
        X := 0;
        Src := Pointer(Long(Data) + Y * WidthBytes);
        while X < Clone.Levels[0].Width do
        begin
          if (Clone.Levels[0].Width - X > 2) and (Src^ = PByte(Long(Src) + 1)^)
            then
          begin
            B1 := 2;
            B2 := Src^;
            Inc(X, 2);
            Inc(Src, 2);
            while (X < Clone.Levels[0].Width) and (Src^ = B2) and (B1 < 255) do
            begin
              Inc(B1);
              Inc(X);
              Inc(Src);
            end;
            AllocByte^ := B1;
            AllocByte^ := B2;
          end
          else if (Clone.Levels[0].Width - X > 2) and (Src^ <> PByte(Long(Src) +
            1)^) and
            (PByte(Long(Src) + 1)^ = PByte(Long(Src) + 2)^) then
          begin
            AllocByte^ := 1;
            AllocByte^ := Src^;
            Inc(Src);
            Inc(X);
          end
          else
          begin
            if (Clone.Levels[0].Width - X < 4) then
            begin
              if Clone.Levels[0].Width - X = 2 then
              begin
                AllocByte^ := 1;
                AllocByte^ := Src^;
                Inc(Src);
                AllocByte^ := 1;
                AllocByte^ := Src^;
                Inc(Src);
                Inc(X, 2);
              end
              else
              begin
                AllocByte^ := 1;
                AllocByte^ := Src^;
                Inc(Src);
                Inc(X);
              end;
            end
            else
            begin
              PB1 := Size;
              AllocByte;
              PB2 := Size;
              AllocByte;
              B1 := 0;
              B2 := 3;
              Inc(X, 3);
              AllocByte^ := Src^;
              Inc(Src);
              AllocByte^ := Src^;
              Inc(Src);
              AllocByte^ := Src^;
              Inc(Src);
              while (X < Clone.Levels[0].Width) and (B2 < 255) do
              begin
                if (Clone.Levels[0].Width - X > 3) and (Src^ = PByte(Long(Src) +
                  1)^) and
                  (Src^ = PByte(Long(Src) + 2)^) and (Src^ = PByte(Long(Src) +
                  3)^) then
                  Break;
                AllocByte^ := Src^;
                Inc(Src);
                Inc(B2);
                Inc(X);
              end;
              PByte(Long(Buff) + PB1)^ := B1;
              PByte(Long(Buff) + PB2)^ := B2;
            end;
          end;
          if Size and 1 = 1 then
            AllocByte;
        end;
        AllocByte^ := 0;
        AllocByte^ := 0;
      end;
      AllocByte^ := 0;
      AllocByte^ := 1;
    finally
      OutPut.WriteBuffer(Buff^, Size);
      FreeMemNil(Buff);
    end;
  end;


begin
  Result := false;
  Assert(Output <> nil, ErrString(Self, 'EncodeStream',
    'Output stream is nil!'));

  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  PrepareImage(Image,Clone);
  StartPos := OutPut.Position;

  with OutPut, Clone do
  try
    try

      WidthBytes := (((Levels[0].Width * BitCount) + 31) shr 5) * 4;

      // other fields will be filled later - we don't know all values now
      FillChar(BF, SizeOf(BF), 0);
      BF.ID := 19778;

      WriteBuffer(BF, SizeOF(BF));

      // other fields will be filled later - we don't know all values now
      FillChar(BI, SizeOf(BI), 0);
      BI.Size := SizeOf(BI);
      BI.Width := Levels[0].Width;
      BI.Height := Levels[0].Height;
      BI.Planes := 1;

      {DONE -o Galfar : Convert nonstandard images to standard ones when saving.}

      BI.BitCount := BitCount;

      if (BitCount = 8) and (FUseRLE) then
        BI.Compression := BI_RLE8
      else if (BitCount = 16) then
        BI.Compression := BI_BITFIELDS
      else
        BI.Compression := BI_RGB;

      if Clone.Palette <> nil then
      begin
        BI.ClrUsed := PalSize;
        BI.ClrImportant := PalSize;
      end;

      WriteBuffer(BI, SizeOf(BI));

       // write mask info
      if BI.Compression = BI_BITFIELDS then
        with ImageFormatToPF(Clone.ImageFormat) do
        begin
          LocalPF.RBitMask := RBitMask;
          LocalPF.GBitMask := GBitMask;
          LocalPF.BBitMask := BBitMask;
          WriteBuffer(LocalPF, SizeOf(LocalPF));
        end;

      // write palette
      if Palette <> nil then
        WriteBuffer(Palette^, PalSize * SizeOf(TARGBRec));

      Off := Position;

      AddPadBytes(Clone, Data, DSize);

      case BI.Compression of
        BI_RGB, BI_BITFIELDS: SaveRGB;
        BI_RLE8: SaveRLE8;
      end;

      FreeMemNil(Data);

      FSize := Position - StartPos;

       // rewrite header with new values
      Position := StartPos;
      BF.Size := FSize;
      BF.Offset := Off;
      BI.SizeImage := FSize - Off;

      WriteBuffer(BF, SizeOf(BF));
      WriteBuffer(BI, SizeOf(BI));

      Position := StartPos + FSize;

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving bitmap data! Image can be corrupted.'),EImage);
      Position := StartPos;
    end;
  finally

  end;

end;

function TBmpModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TBmpModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  BF: TBitmapFileHeader;
  BI: TBitmapInfoHeader;
  BC: TBitmapCoreHeader;
  IsOS2: boolean;
  StartPos: Int64;
  LocalPF: TLocalPixelFormat;
  PalRGB: PPalRGBArray;
  I: Long;
  WidthBytes: Long;


  procedure Convert1To8(var Image: TImageData);
  const
    Mask1: array[0..7] of Byte = ($80, $40, $20, $10, $08, $04, $02, $01);
    Shift1: array[0..7] of Byte = (7, 6, 5, 4, 3, 2, 1, 0);

  var
    Data: Pointer;
    X, Y: Long;
    WB: Long;
  begin
    with Image, Image.Levels[0] do
    begin
      if (BitCount <> 1) or (Bits = nil) then
        Exit;

      GetMem(Data, Width * Height);
      WB := (((Width * BitCount) + 31) shr 5) * 4;
      try
        if PalSize > 0 then
        begin
          PalSize := 256;
          ReallocMem(Palette, PalSize * SizeOf(TARGBRec));
        end;
        for X := 0 to Width - 1 do
          for Y := 0 to Height - 1 do
            PByte(Long(Data) + Y * Width + X)^ :=
              (PByteArray(Integer(Bits) + Y * WB)[X shr 3] and Mask1[X and 7])
                shr
              Shift1[X and 7];
      finally
        FreeMemNil(Bits);
        Bits := Data;
        if PalSize > 0 then
          ImageFormat := ifIndex8
        else
          ImageFormat := ifIntensity8;
        BitCount := 8;
        Size := Width * Height;
      end;
    end;
  end;


  procedure Convert4To8(var Image: TImageData);
  const
    Mask4: array[0..1] of Byte = ($F0, $0F);
    Shift4: array[0..1] of Byte = (4, 0);

  var
    Data: Pointer;
    X, Y: Long;
    WB: Long;
  begin
    with Image, Image.Levels[0] do
    begin
      if (BitCount <> 4) or (Bits = nil) then
        Exit;

      GetMem(Data, Width * Height);

      WB := (((Width * BitCount) + 31) shr 5) * 4;
      try
        if PalSize > 0 then
        begin
          PalSize := 256;
          ReallocMem(Palette, PalSize * SizeOf(TARGBRec));
        end;
        for X := 0 to Width - 1 do
          for Y := 0 to Height - 1 do
            PByte(Long(Data) + Y * Width + X)^ :=
              (PByteArray(Integer(Bits) + Y * WB)[X shr 1] and Mask4[X and 1])
                shr
              Shift4[X and 1];
      finally
        FreeMemNil(Bits);
        Bits := Data;
        if PalSize > 0 then
          ImageFormat := ifIndex8
        else
          ImageFormat := ifIntensity8;
        BitCount := 8;
        Size := Width * Height;
      end;
    end;
  end;


  procedure LoadRGB;
  var
    I: Long;
  begin
    with Image.Levels[0] do
    begin
      GetMem(Bits, Size);
      // if Height is <0 then image data are stored non-flipped (like in OGL+D3D)
      // but default in windows is flipped so if Height is positive we must flip it
      if BI.Height < 0 then
        Input.ReadBuffer(Bits^, Size)
      else
        for I := Height - 1 downto 0 do
          Input.ReadBuffer(Pointer(Long(Bits) + I * WidthBytes)^,
            WidthBytes);
    end;
  end;


  procedure LoadRLE4;
  var
    B1, B2, C: Byte;
    Data, Dest, Src, P: PByte;
    X, Y, I: Long;
    BW: Long;
  begin
    with Image.Levels[0] do
    begin
      BW := (((Width * 4) + 31) shr 5) * 4;
      Size := Height * BW;
      GetMem(Data, BI.SizeImage);
      GetMem(Bits, Size);
      try
        Input.ReadBuffer(Data^, BI.SizeImage);
        X := 0;
        Y := 0;
        // this line must be here because if you directly use Data
        // in processing then FreeMemNilNil in finally section will raise exception
        Src := Data;
        while True do
        begin
          B1 := Src^;
          Inc(Src);
          B2 := Src^;
          Inc(Src);
          if B1 = 0 then
          begin
            case B2 of
              0:
                begin //  End of line
                  X := 0;
                  Inc(Y);
                end;
              1: Break; //  End of bitmap
              2:
                begin //  Difference of coordinates
                  Inc(X, B1);
                  Inc(Y, B2);
                  Inc(Src, 2);
                end;
            else
              Dest := Pointer(Long(Bits) + Y * BW);
              C := 0;
              for i := 0 to B2 - 1 do
              begin
                if i and 1 = 0 then
                begin
                  C := Src^;
                  Inc(Src);
                end
                else
                begin
                  C := C shl 4;
                end;
                P := Pointer(Long(Dest) + X shr 1);
                if X and 1 = 0 then
                  P^ := (P^ and $0F) or (C and $F0)
                else
                  P^ := (P^ and $F0) or ((C and $F0) shr 4);
                Inc(X);
              end;
            end;
          end
          else
          begin
            Dest := Pointer(Long(Bits) + Y * BW);
            for I := 0 to B1 - 1 do
            begin
              P := Pointer(Long(Dest) + X shr 1);
              if X and 1 = 0 then
                P^ := (P^ and $0F) or (B2 and $F0)
              else
                P^ := (P^ and $F0) or ((B2 and $F0) shr 4);
              Inc(X);
              // Swap nibble
              B2 := (B2 shr 4) or (B2 shl 4);
            end;
          end;
          //  Word arrangement
          Inc(Src, Long(Src) and 1);
        end;
      finally
        FreeMemNil(Data);
      end;
    end;
  end;


  procedure LoadRLE8;
  var
    B1, B2: Byte;
    Data, Dest, Src: PByte;
    X, Y, BW: Long;
  begin
    with Image.Levels[0] do
    begin
      BW := (((Width * 8) + 31) shr 5) * 4;
      Size := Height * BW;
      GetMem(Data, BI.SizeImage);
      GetMem(Bits, Size);
      try
        Input.ReadBuffer(Data^, BI.SizeImage);
        Dest := Bits;
        Src := Data;
        X := 0;
        Y := 0;
        while True do
        begin
          B1 := Src^;
          Inc(Src);
          B2 := Src^;
          Inc(Src);
          if B1 = 0 then
          begin
            case B2 of
              0:
                begin //  End of line
                  X := 0;
                  Inc(Y);
                  Dest := Pointer(Long(Bits) + Y * BW + X);
                end;
              1: Break; //  End of bitmap
              2:
                begin //  Difference of coordinates
                  Inc(X, B1);
                  Inc(Y, B2);
                  Inc(Src, 2);
                  Dest := Pointer(Long(Bits) + Y * BW + X);
                end;
            else
              Move(Src^, Dest^, B2);
              Inc(Dest, B2);
              Inc(Src, B2);
            end;
          end
          else
          begin
            FillChar(Dest^, B1, B2);
            Inc(Dest, B1);
          end;
          //  Word arrangement
          Inc(Src, Longint(Src) and 1);
        end;
      finally
        FreeMemNil(Data);
      end;
    end;
  end;


begin
  Result := false;
  FreeImageSafe(Image);
  Assert(TestFormat(Input), ErrString(Self, 'DecodeStream',
    'Stream image data not in BMP/DIB format or corrupted!'));

  StartPos := Input.Position;
  with Input do
  try
    try
      ReadBuffer(BF, SizeOf(BF));
      ReadBuffer(BI.Size, SizeOf(BI.Size));

      IsOS2 := BI.Size = SizeOf(TBitmapCoreHeader);

      // Bitmap Info reading
      if IsOS2 then
      begin
        // OS/2 type bitmap
        // Reads info header without 4 already read bytes
        ReadBuffer(Pointer(Long(@BC) + 4)^, SizeOf(TBitmapCoreHeader) - 4);

        with BI do
        begin
          ClrUsed := 0;
          Compression := BI_RGB;
          BitCount := BC.BitCount;
          Height := BC.Height;
          Width := BC.Width;
        end;
      end
      else
        //  Windows type bitmap
        ReadBuffer(Pointer(Long(@BI) + 4)^, SizeOf(TBitmapInfoHeader) - 4);

      AddLevel(Image);
      Image.Levels[0].Width := BI.Width;
      Image.Levels[0].Height := Abs(BI.Height);
      Image.BitCount := BI.BitCount;
      Image.Levels[0].Size := BI.SizeImage;

      WidthBytes := (((Image.Levels[0].Width * Image.BitCount) + 31) shr 5) * 4;

      if Image.Levels[0].Size = 0 then
        Image.Levels[0].Size := Image.Levels[0].Height * WidthBytes;


      //  Bit mask reading.
      if BI.Compression = BI_BITFIELDS then
        ReadBuffer(LocalPF, SizeOf(LocalPF));

      case BI.BitCount of
        1, 4, 8: Image.ImageFormat := ifIndex8;
        16: if LocalPF.RBitMask = 63488 then
            Image.ImageFormat := ifR5G6B5
            else
            if LocalPF.RBitMask = 3480 then
              Image.ImageFormat := ifA4R4G4B4
            else
            Image.ImageFormat := ifX1R5G5B5;
        24: Image.ImageFormat := ifR8G8B8;
        32: Image.ImageFormat := ifX8R8G8B8;
      end;

        //  Palette settings and reading
      Image.PalSize := 0;
      Image.Palette := nil;

      if Image.BitCount <= 8 then
        if IsOS2 then
        begin
        //  OS/2 type
          Image.PalSize := Round(Power2(Image.BitCount));
          GetMem(Image.Palette, Image.PalSize * SizeOf(TARGBRec));

          GetMem(PalRGB, Image.PalSize * SizeOf(TRGBRec));
          ReadBuffer(PalRGB^, Image.PalSize * SizeOf(TRGBRec));
          for I := 0 to Image.PalSize - 1 do
          begin
            with PalRGB[I] do
              Image.Palette[I] := ARGBRec(ARGB($FF, R, G, B));
          end;
          FreeMemNil(PalRGB);
        end
        else
        begin
         //  Windows type  }
          Image.PalSize := BI.ClrImportant;
          GetMem(Image.Palette, Image.PalSize * SizeOf(TARGBRec));
          ReadBuffer(Image.Palette^, Image.PalSize * SizeOf(TARGBRec));
        end;

      case BI.Compression of
        BI_RGB: LoadRGB;
        BI_RLE4: LoadRLE4;
        BI_RLE8: LoadRLE8;
        BI_BITFIELDS: LoadRGB;
      end;

      // 1 and 4 bpp images are supported only for loading which is now
      // so we now convert them to 8bpp.
      //must be called before RemovePadBytes or remove will fail
      case Image.BitCount of
        1: Convert1To8(Image);
        4: Convert4To8(Image);
      end;

      RemovePadBytes(Image);

      if BI.Compression in [BI_RLE4, BI_RLE8] then
        if BI.Height > 0 then
          FlipYImage(Image);

      Position := StartPos + BF.Size;

    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading bitmap data! Image can be corrupted.'),EImage);
      Position := StartPos;
    end;
  finally

  end;
  Result := true;
end;


const
  MinSize = SizeOf(TBitmapFileHeader) + SizeOf(TBitmapCoreHeader);


function TBmpModule.TestFormat(Data: Pointer; Size: Long): boolean;
begin
  Result := false;
  if (Data = nil) or (Size < MinSize) then
    Exit;

  if CompareMem(Data, @BMPMagic, 2) then
    Result := true;
end;

function TBmpModule.TestFormat(Input: TStream): boolean;
var
  ID: array[0..1] of Char;
  OldPos: Int64;
begin
  Result := false;
  if Input = nil then
    Exit;
  if Input.Size < MinSize then
    Exit;

  OldPos := Input.Position;
  Input.ReadBuffer(ID, SizeOf(ID));
  Input.Position := OldPos;

  if CompareMem(@ID, @BMPMagic, SizeOf(ID)) then
    Result := true;
end;


initialization

  BmpModuleInfo:=GetDefaultIntModuleInfo;
  BmpModuleInfo.Name:=sBmpName;
  BmpModuleInfo.Description:=sBmpDesc;
  BmpModuleInfo.IDCode:=sBmpCode;
  RegisterIntModule(@BmpModuleInfo,TBmpModule);

end.
 