{*************************************************************************}
{*   Dagger                                                              *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images\Daggerfall\Dagger.pas                                *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 2nd March 2003  (02.03.2003)                               }
{*************************************************************************}


{This unit contains modules for reading images/anims/textures from
The Elder Scrolls Chapter 2: Daggerfall. It can handle all! *.img,*.cif,texture.*
files in this game (Empty files (which are! in Dag) are of course skipped).
However, it can handle images from more older Bethesda's games
(TES 1: Arena, Terminator: Future Shock + Skynet,...).}


unit Dagger;


interface


uses
  Sys, Objects, Storage, Images;


type


  EDaggerfall = class(EImage);


{*************************************************************************}
{                        Daggerfall modules                               }
{*************************************************************************}


  {Interface for TDagerfallModule.}
  IDaggerfallModule = interface(IImageModule)
    ['{EBE94F7C-4322-461E-B28E-4B56373750D0}']

    function GetPalette: TPalette;
    procedure SetPalette(const Value: TPalette);

    procedure ConvertToDagg(var Image: TImageData);

    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean;
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;

    function LoadPal(Data: Pointer; Size: Long): boolean; overload;
    function LoadPal(Input: TStream): boolean; overload;

    function TestFormat(Data: Pointer; Size: Long): boolean; overload;
    function TestFormat(Input: TStream): boolean; overload;

    property Palette: TPalette read GetPalette write SetPalette;

  end;

  {Abstract class for all Daggerfall modules.
  They require loaded palette (*.pal;*.col in Dag\Arena2 directory).
  Now default palette pal.pal is contained in this unit - it works
  well for 99% of Daggerfall graphics.
  They do not support continuous streams - exact size must be known
  before loading (because some files are without headers).}
  TDaggerfallModule = class(TImageModule, IDaggerfallModule)
  protected

    FPalette: TPalette;

    function GetPalette: TPalette;

    procedure SetPalette(const Value: TPalette);

    {Here convert from image formats not supported by encoder }
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      override;


  public

    constructor Create; override;
    destructor Destroy; override;

    {Converts image to 8bit indexed using Daggerfall loaded (or default) palette.}
    procedure ConvertToDagg(var Image: TImageData);

    {Loads palette and sets active pal to this value.
    Size must be total size of palette.}
    function LoadPal(Data: Pointer; Size: Long): boolean; overload;

    {Loads palette and sets active pal to this value.
    Stream must contain only! palette, nothing more.}
    function LoadPal(Input: TStream): boolean; overload;

    {This returns always false. Dagg images has no meaningfull format IDs.}
    function TestFormat(Data: Pointer; Size: Long): boolean; overload; override;

    {This returns always false. Dagg images has no meaningfull format IDs.}
    function TestFormat(Input: TStream): boolean; overload; override;


    {Active palette used in loading}
    property Palette: TPalette read GetPalette write SetPalette;

  published

  end;


  {This module proveides loading/saving of images in IMG format.
  This format is simple non-compressed 8bit indexed.
  But some files are without header!  so exact file size must be known.
  This is why Stream functions must contain only one IMG file.}
  TImgModule = class(TDaggerfallModule)
  public

    constructor Create; override;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;


    {This func converts Data read from image file to valid Image Data
    structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func converts Input stream read from image file to valid Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;

  published

  end;


  {This module proveides loading/saving of images in CIF format.
  This format is simple non-compressed 8bit indexed containing more frames.
  But some files are without header and some are RLEd. So exact file size must be known.
  This is why Stream functions must contain only one CIF file.}
  TCifModule = class(TDaggerfallModule)
  public

    constructor Create; override;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This func takes valid image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;


    {This func converts Data read from image file to valid Image Data
    structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func converts Input stream read from image file to valid Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;

  published

  end;


  {This module proveides loading of Daggerfall textures.
  This format is very complicated ( more images with subimages,
  non-standard RLE, many unknowns) so module supports only loading.}
  TTextureModule = class(TDaggerfallModule)
  protected

    FLastName: string;

  public

    constructor Create; override;

    {This causes exception}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This causes exception}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;


    {This func converts Data read from image file to valid Image Data
    structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func converts Input stream read from image file to valid Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;


    {Stores name of last loaded texture.}
    property LastTextureName: string read FLastName;

  published

  end;


{*************************************************************************}
{                         Some other type/consts                          }
{*************************************************************************}


const

  sImgName = 'Img Module';
  sImgDesc =
    'Provides Daggerfall Img (*.img) image format encoding/decoding';
  sImgCode: TIDCode = 'IMG-DIMG';

  sCifName = 'Cif Module';
  sCifDesc =
    'Provides Daggerfall Cif (*.cif) image format encoding/decoding';
  sCifCode: TIDCode = 'IMG-DCIF';

  sTexName = 'Texture Module';
  sTexDesc =
    'Provides Daggerfall Texture (texture.*) image format decoding';
  sTexCode: TIDCode = 'IMG-DTEX';


type

  {Header for Img files}
  TImgHeader = packed record
    XOff: Word;
    YOff: Word;
    Width: Word;
    Height: Word;
    Unk: Word; // Might indicate compressed data or not
    ImageSize: Word; // Size of Image data (but not always)
  end;


  {Header for Cif Group files}
  TCifGroup = packed record
    Width: Word;
    Height: Word;
    XOff: Word;
    YOff: Word;
    Unk: Word;
    ImageSize: Word; // Size of Image data (but not always)
    Offsets: array[0..31] of Word; //offsets from beginning of header to
      // image datas. Last offset points to next group header
  end;


  {Main header for texture}
  TTexHeader = packed record
    ImgCount: Word; // number of images in texture
    TexName: array[0..23] of Char; // name of texture
  end;


  {Offset list  for texture}
  TOffset = packed record
    Type1: Word; // ??
    HdrOffset: ULong; // contains offsetof Img header from the origin
     // of the file
    Type2: Word; // ??
    Unk: ULong; // ranges from 0 to 4 (0 in 90%)
    Null1: ULong; // always 0
    Null2: ULong; // always 0
  end;


  TOffsetList = array[0..0] of TOffset;
  POffsetList = ^TOffsetList;


  {Image header for texture}
  TTexImgHeader = packed record
    XOff: Word;
    YOff: Word;
    Width: Word;
    Height: Word;
    Unk1: Word; // $0108 = Image has subimages which are RLE compressed data.
     //	$1108 = Image has RLE type compressed data with a row offset section
     //before the single image data.
    ImageSize: Long; // image size (including header)
    ImageOff: Long; // pointer to start of image data from this header
    Unk2: Word; //  $0000 = Image has subimages in special compressed format.
    //$00C0 = Usual value, regular single image.
    //NonZero = Regular single image.Unknown what the differences indicate
    SubImages: Word; // number of subimages (1 = single image)
    Unk3: Long;
    Unk4: Word;
  end;


var

  ImgModuleInfo:TModuleInfo;
  CifModuleInfo:TModuleInfo;
  TexModuleInfo:TModuleInfo;


{Decodes RLE used in IMG and CIF files. OutSize must be known before calling!
(use Out = Width * Height)}
procedure DagRLEDecode(InData: Pointer; InSize, OutSize: Long; out OutData:
  Pointer);

{Deletes non-valid chars from texture name.}
function RepairName(const S: string): string;


const


  {This is default Daggerfall's palette (C:\Dagger\Arena2\pal.pal).
  Every Daggerfall module loads this pal in constructor.}
  DefaultPal: TPalRGB = (
    (B: 0; G: 0; R: 0), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255;
    G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255), (B: 255; G: 0; R: 255),
    (B: 255; G: 0; R: 255)
    , (B: 244; G: 202; R: 167), (B: 227; G: 180; R: 144), (B: 207; G: 152; R:
    118), (B: 193; G: 133; R: 100)
    , (B: 180; G: 113; R: 80), (B: 165; G: 100; R: 70), (B: 152; G: 93; R: 63),
    (B: 140; G: 86; R: 55)
    , (B: 129; G: 79; R: 48), (B: 122; G: 75; R: 43), (B: 112; G: 70; R: 40),
    (B: 103; G: 64; R: 39)
    , (B: 91; G: 67; R: 38), (B: 79; G: 63; R: 43), (B: 66; G: 54; R: 41), (B:
    54; G: 50; R: 40)
    , (B: 232; G: 196; R: 196), (B: 220; G: 177; R: 177), (B: 204; G: 157; R:
    157), (B: 188; G: 138; R: 138)
    , (B: 175; G: 122; R: 122), (B: 155; G: 105; R: 106), (B: 143; G: 94; R:
    97), (B: 126; G: 81; R: 89)
    , (B: 109; G: 72; R: 88), (B: 101; G: 68; R: 85), (B: 86; G: 61; R: 77), (B:
    75; G: 55; R: 71)
    , (B: 67; G: 51; R: 63), (B: 63; G: 47; R: 56), (B: 56; G: 45; R: 52), (B:
    46; G: 44; R: 46)
    , (B: 245; G: 212; R: 172), (B: 229; G: 193; R: 150), (B: 213; G: 174; R:
    128), (B: 196; G: 154; R: 105)
    , (B: 183; G: 140; R: 88), (B: 173; G: 127; R: 78), (B: 160; G: 118; R: 74),
    (B: 151; G: 110; R: 69)
    , (B: 134; G: 103; R: 65), (B: 123; G: 92; R: 60), (B: 109; G: 85; R: 54),
    (B: 96; G: 76; R: 51)
    , (B: 83; G: 71; R: 44), (B: 69; G: 63; R: 42), (B: 61; G: 54; R: 38), (B:
    50; G: 45; R: 34)
    , (B: 205; G: 205; R: 224), (B: 188; G: 188; R: 199), (B: 165; G: 165; R:
    174), (B: 145; G: 145; R: 159)
    , (B: 135; G: 135; R: 149), (B: 122; G: 122; R: 137), (B: 114; G: 114; R:
    127), (B: 103; G: 103; R: 116)
    , (B: 94; G: 94; R: 109), (B: 85; G: 85; R: 96), (B: 75; G: 75; R: 85), (B:
    68; G: 68; R: 80)
    , (B: 61; G: 61; R: 67), (B: 53; G: 53; R: 59), (B: 48; G: 48; R: 50), (B:
    44; G: 44; R: 45)
    , (B: 176; G: 205; R: 255), (B: 147; G: 185; R: 244), (B: 123; G: 164; R:
    230), (B: 104; G: 152; R: 217)
    , (B: 87; G: 137; R: 205), (B: 68; G: 124; R: 192), (B: 68; G: 112; R: 179),
    (B: 62; G: 105; R: 167)
    , (B: 55; G: 97; R: 154), (B: 49; G: 90; R: 142), (B: 45; G: 82; R: 122),
    (B: 51; G: 77; R: 102)
    , (B: 52; G: 69; R: 87), (B: 50; G: 62; R: 73), (B: 47; G: 59; R: 60), (B:
    44; G: 48; R: 49)
    , (B: 220; G: 220; R: 220), (B: 197; G: 197; R: 197), (B: 185; G: 185; R:
    185), (B: 174; G: 174; R: 174)
    , (B: 162; G: 162; R: 162), (B: 147; G: 147; R: 147), (B: 132; G: 132; R:
    132), (B: 119; G: 119; R: 119)
    , (B: 110; G: 110; R: 110), (B: 99; G: 99; R: 99), (B: 87; G: 87; R: 87),
    (B: 78; G: 78; R: 78)
    , (B: 67; G: 67; R: 67), (B: 58; G: 58; R: 58), (B: 51; G: 51; R: 51), (B:
    44; G: 44; R: 44)
    , (B: 182; G: 218; R: 227), (B: 158; G: 202; R: 202), (B: 134; G: 187; R:
    187), (B: 109; G: 170; R: 170)
    , (B: 87; G: 154; R: 154), (B: 77; G: 142; R: 142), (B: 70; G: 135; R: 135),
    (B: 62; G: 124; R: 124)
    , (B: 54; G: 112; R: 112), (B: 46; G: 103; R: 103), (B: 39; G: 91; R: 91),
    (B: 40; G: 83; R: 83)
    , (B: 45; G: 72; R: 72), (B: 47; G: 63; R: 63), (B: 50; G: 55; R: 55), (B:
    45; G: 48; R: 48)
    , (B: 255; G: 246; R: 103), (B: 241; G: 238; R: 45), (B: 226; G: 220; R: 0),
    (B: 212; G: 203; R: 0)
    , (B: 197; G: 185; R: 0), (B: 183; G: 168; R: 0), (B: 168; G: 150; R: 0),
    (B: 154; G: 133; R: 0)
    , (B: 139; G: 115; R: 0), (B: 127; G: 106; R: 4), (B: 116; G: 97; R: 7), (B:
    104; G: 87; R: 11)
    , (B: 93; G: 78; R: 14), (B: 81; G: 69; R: 18), (B: 69; G: 60; R: 21), (B:
    58; G: 51; R: 25)
    , (B: 202; G: 221; R: 196), (B: 175; G: 200; R: 168), (B: 148; G: 176; R:
    141), (B: 123; G: 156; R: 118)
    , (B: 107; G: 144; R: 109), (B: 93; G: 130; R: 94), (B: 82; G: 116; R: 86),
    (B: 77; G: 110; R: 78)
    , (B: 68; G: 99; R: 67), (B: 61; G: 89; R: 53), (B: 52; G: 77; R: 45), (B:
    46; G: 68; R: 37)
    , (B: 39; G: 60; R: 39), (B: 30; G: 55; R: 30), (B: 34; G: 51; R: 34), (B:
    40; G: 47; R: 40)
    , (B: 179; G: 107; R: 83), (B: 175; G: 95; R: 75), (B: 175; G: 87; R: 67),
    (B: 163; G: 79; R: 59)
    , (B: 155; G: 75; R: 51), (B: 147; G: 71; R: 47), (B: 155; G: 91; R: 47),
    (B: 139; G: 83; R: 43)
    , (B: 127; G: 75; R: 39), (B: 115; G: 67; R: 35), (B: 99; G: 63; R: 31), (B:
    87; G: 55; R: 27)
    , (B: 75; G: 47; R: 23), (B: 59; G: 39; R: 19), (B: 47; G: 31; R: 15), (B:
    35; G: 23; R: 11)
    , (B: 216; G: 227; R: 162), (B: 185; G: 205; R: 127), (B: 159; G: 183; R:
    101), (B: 130; G: 162; R: 77)
    , (B: 109; G: 146; R: 66), (B: 101; G: 137; R: 60), (B: 92; G: 127; R: 54),
    (B: 84; G: 118; R: 48)
    , (B: 76; G: 108; R: 42), (B: 65; G: 98; R: 37), (B: 53; G: 87; R: 34), (B:
    51; G: 75; R: 35)
    , (B: 45; G: 64; R: 37), (B: 43; G: 56; R: 39), (B: 38; G: 51; R: 40), (B:
    43; G: 46; R: 45)
    , (B: 179; G: 115; R: 79), (B: 175; G: 111; R: 75), (B: 171; G: 107; R: 71),
    (B: 167; G: 103; R: 67)
    , (B: 159; G: 99; R: 63), (B: 155; G: 95; R: 59), (B: 151; G: 91; R: 55),
    (B: 143; G: 87; R: 51)
    , (B: 40; G: 40; R: 40), (B: 38; G: 38; R: 38), (B: 35; G: 35; R: 35), (B:
    31; G: 31; R: 31)
    , (B: 27; G: 27; R: 27), (B: 23; G: 23; R: 23), (B: 19; G: 19; R: 19), (B:
    15; G: 15; R: 15)
    , (B: 254; G: 255; R: 199), (B: 254; G: 245; R: 185), (B: 254; G: 235; R:
    170), (B: 254; G: 225; R: 156)
    , (B: 255; G: 215; R: 141), (B: 255; G: 205; R: 127), (B: 255; G: 195; R:
    112), (B: 255; G: 185; R: 98)
    , (B: 255; G: 175; R: 83), (B: 241; G: 167; R: 54), (B: 234; G: 155; R: 50),
    (B: 226; G: 143; R: 46)
    , (B: 219; G: 131; R: 43), (B: 212; G: 119; R: 39), (B: 205; G: 107; R: 35),
    (B: 198; G: 95; R: 31)
    , (B: 190; G: 84; R: 27), (B: 183; G: 72; R: 23), (B: 176; G: 60; R: 19),
    (B: 169; G: 48; R: 15)
    , (B: 162; G: 36; R: 12), (B: 154; G: 24; R: 8), (B: 147; G: 12; R: 4), (B:
    130; G: 22; R: 0)
    , (B: 111; G: 34; R: 0), (B: 102; G: 33; R: 1), (B: 92; G: 33; R: 3), (B:
    83; G: 32; R: 10)
    , (B: 74; G: 39; R: 27), (B: 65; G: 41; R: 33), (B: 57; G: 43; R: 39), (B:
    48; G: 45; R: 45));


implementation


procedure DagRLEDecode(InData: Pointer; InSize, OutSize: Long; out OutData:
  Pointer);
var
  I, Pos, CByte: Long;
  Rle, B: Byte;

begin
  GetMem(OutData, OutSize);
  Pos := 0;
  CByte := 0;
  while Pos < OutSize do
  begin
    Rle := PByte(Long(InData) + CByte)^;
    CByte := CByte + 1;
    if Rle < 128 then
    begin
      Rle := Rle + 1;
      Move(PByte(Long(InData) + CByte)^, PByte(Long(OutData) + Pos)^, Rle);
      CByte := CByte + Rle;
      Pos := Pos + Rle;
    end
    else
    begin
      Rle := Rle - 127;
      B := PByte(Long(InData) + CByte)^;
      CByte := CByte + 1;
      for I := 0 to Rle - 1 do
      begin
        PByte(Long(OutData) + Pos)^ := B;
        Pos := Pos + 1;
      end;
    end;
  end;
end;

function RepairName(const S: string): string;
var
  I: Long;
  First: boolean;
begin
  I := 1;
  Result := S;
  First := false;
  while I <= Length(Result) do
    if (Ord(Result[I]) < 32) or
      ((Ord(Result[I]) = 32) and (not First)) then
      Delete(Result, I, 1)
    else
    begin
      Inc(I);
      First := true;
    end;
end;


{*************************************************************************}
{                         TDaggerfallModule Methods                       }
{*************************************************************************}


constructor TDaggerfallModule.Create;
begin
  inherited Create;
  FPalette := RGBPalToARGBPal(DefaultPal);
end;

destructor TDaggerfallModule.Destroy;
begin

  inherited Destroy;
end;

function TDaggerfallModule.GetPalette: TPalette;
begin
  Result := FPalette;
end;

procedure TDaggerfallModule.SetPalette(const Value: TPalette);
begin
  FPalette := Value;
end;

procedure TDaggerfallModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  inherited PrepareImage(Image, Clone);
  // Daggerfall supports 8bit indexed images only -> convert all to 8bit
  //   using default Dagg palette
  ConvertToDagg(Clone);
end;

procedure TDaggerfallModule.ConvertToDagg(var Image: TImageData);
var
  Data: Pointer;
  I, X, Y: Long;
  B, BS: PByte;
  C: TARGB;

  function GetColor: TARGB;
  var
    PF: TPixelFormat;
  begin
    Result:=0;
    with Image, Image.Levels[I] do
      case BitCount of
        8: case ImageFormat of
            ifIndex8: Result := Palette[BS^].Color;
            ifIntensity8: Result := ($FF shl 24) or (BS^ shl 16) or (BS^ shl 8)
              or BS^;
            Low(if8bitARGBPF) .. High(if8bitARGBPF):
              begin
                PF := ImageFormatToPF(ImageFormat);
                Result := PFGetARGB(PF, BS^);
              end;
          end;
        16:
          begin
            PF := ImageFormatToPF(ImageFormat);
            Result := PFGetARGB(PF, PWord(Pointer(BS))^);
          end;
        24: with PRGBRec(Pointer(BS))^ do
            Result := (R shl 16) or (G shl 8) or B;
        32: Result := PULong(Pointer(BS))^;
      end;
  end;

begin
  Image.PalSize := 256;
  if Image.Palette = nil then
    GetMem(Image.Palette, Image.PalSize * SizeOf(TARGB))
  else
    ReallocMem(Image.Palette, Image.PalSize * SizeOf(TARGB));
  Move(FPalette, Image.Palette^, Image.PalSize * SizeOf(TARGB));
  PalSwapChannels(Image.Palette, Image.PalSize);

  for I := 0 to Image.LevelCount - 1 do
    with Image, Image.Levels[I] do
    begin
      GetMem(Data, Width * Height);
      Size := Width * Height;
      B := Data;
      BS := Bits;
      for X := 0 to Width - 1 do
        for Y := 0 to Height - 1 do
        begin
          C := GetColor;
          B^ := ClosestColorInPal(Palette, 256, C);
          Inc(B);
          Inc(BS, BitCount shr 3);
        end;
      FreeMemNil(Bits);
      Bits := Data;
    end;
  Image.BitCount := 8;
  Image.ImageFormat := ifIndex8;

  // Img and Cif formats can save max 65536 bytes big images in standard format
  //   so we must resize (Texture(and nonstandard Img and Cif) files can
  //   store larger images, but we are not saving them here)
  {DONE -o Galfar : Add code here}
  for I := 0 to Image.LevelCount - 1 do
    while Image.Levels[I].Size > High(Word) do
      ResizePixels(Image.Levels[I].Width div 2, Image.Levels[I].Height div 2,
        Image, I);
end;

function TDaggerfallModule.LoadPal(Data: Pointer; Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls LoadPal(Stream) function rather that try to parse raw pointer
    Result := LoadPal(MS);
  finally
    MS.Free;
  end;
end;

function TDaggerfallModule.LoadPal(Input: TStream): boolean;
var
  Pal: TPalRGB;
  I: Int64;
begin
  Result := false;
  Assert((Input.Size = 768) or (Input.Size = 776), ErrString(Self,
    'LoadPal',
    'This is not valid Daggerfall palette (must be 768 or 776 bytes long)!'));
  try
    try
      if Input.Size = 776 then
        Input.ReadBuffer(I, SizeOf(I));
      Input.ReadBuffer(Pal, SizeOf(Pal));
      FPalette := RGBPalToARGBPal(Pal, $FF);
    except
      RaiseError(ErrString(Self, 'LoadPal',
        'Error when loading palette!'), EDaggerfall);
    end;
  finally
    Input.Position := 0;
  end;
  Result := true;
end;

function TDaggerfallModule.TestFormat(Data: Pointer; Size: Long): boolean;
begin
  Result := true;
end;

function TDaggerfallModule.TestFormat(Input: TStream): boolean;
begin
  Result := true;
end;


{*************************************************************************}
{                         TImgModule Methods                              }
{*************************************************************************}


constructor TImgModule.Create;
begin
  inherited Create;
  FInfo := ImgModuleInfo;
end;

function TImgModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TImgModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  Hdr: TImgHeader;
  OldPos: Int64;
  Clone: TImageData;

begin
  Result := false;
  Assert(Output <> nil, ErrString(Self, 'EncodeStream',
    'Output stream is nil!'));

  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  OldPos := OutPut.Position;
  PrepareImage(Image, Clone);

  with OutPut, Clone, Clone.Levels[0] do
  try
    try
      FillChar(Hdr, SizeOf(Hdr), 0);
      Hdr.Width := Width;
      Hdr.Height := Height;
      Hdr.ImageSize := Width * Height;

      WriteBuffer(Hdr, SizeOf(Hdr));
      WriteBuffer(Bits^, Hdr.ImageSize);

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving image data! Image can be corrupted.'), EDaggerfall);
      OutPut.Position := OldPos;
    end;
  finally
    FreeImageSafe(Clone);
  end;
  Result := true;
end;

function TImgModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TImgModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  Hdr: TImgHeader;
  Pal: TPalRGB;
  PalA: TPalette;
  Data: Pointer;
  IsRLE: boolean;
  OldPos: Int64;

  procedure SetDims;

    procedure SetWH(W, H: Long);
    begin
      Image.Levels[0].Width := W;
      Image.Levels[0].Height := H;
      Image.Levels[0].Size := W * H;
    end;


  begin
    // special images, without header
    case Input.Size of
      64: SetWH(8, 8); // Arena file
      90: SetWH(9, 10); // Arena file
      128: SetWH(8, 16); // Arena file
      720: SetWH(9, 80);
      990: SetWH(45, 22);
      1720: SetWH(43, 40);
      2140: SetWH(107, 20);
      2916: SetWH(36, 81);
      3200: SetWH(40, 80);
      3938: SetWH(179, 22);
      4096: SetWH(64, 64); // textures from TES:Arena
      4280: SetWH(107, 40);
      4508: SetWH(322, 14);
      20480: SetWH(320, 64);
      26496: SetWH(207, 128);
      64000: SetWH(320, 200);
      64768: SetWH(320, 200); //this contains also palette -> parsed further
      68800: SetWH(320, 215);
      112128: SetWH(512, 219);
    end;
  end;

begin
  Result := false;
  FreeImageSafe(Image);

  AddLevel(Image);

  OldPos := Input.Position;

  // image has null header or is empty
  if Input.Size <= 12 then
  begin
    NewImage(Image, 8, 8, ifIndex8);
  end
  else

    with Input, Image, Image.Levels[0] do
    try
      try
        BitCount := 8;
        ImageFormat := ifIndex8;
        IsRLE := false;

        ReadBuffer(Hdr, SizeOf(Hdr));
        // solve many img types
        if (Hdr.Width * Hdr.Height = Hdr.ImageSize) and (
          (Hdr.Width * Hdr.Height = Input.Size - SizeOf(Hdr)) or
          (Hdr.Width * Hdr.Height = Input.Size - SizeOf(Hdr) - 768)) then
        begin
          // uncompressed with header
          Width := Hdr.Width;
          Height := Hdr.Height;
          Size := Width * Height;
        end
        else
          if ((Hdr.ImageSize = Input.Size - SizeOf(Hdr)) or
            (Hdr.ImageSize = Input.Size - SizeOf(Hdr) - 768)) and
          (Hdr.Unk <> 0) then
          begin
            // compressed with header
            Width := Hdr.Width;
            Height := Hdr.Height;
            Size := Width * Height;
            IsRLE := true;
          end
          else
          begin
            // uncompressed, no header
            SetDims;
            Position := Position - SizeOf(Hdr);
          end;

        if not IsRLE then
        begin
          // read uncompressed data
          if (Width = 0) or (Height = 0) then
            // file has no header and dims are unknown
            NewImage(Image, 8, 8, ifIndex8)
          else
          begin
            GetMem(Bits, Size);
            ReadBuffer(Bits^, Size);
          end;
        end
        else
          if Hdr.Unk = 2 then
          begin
            // read compressed data Daggerfall
            GetMem(Data, Hdr.ImageSize);
            ReadBuffer(Data^, Hdr.ImageSize);
            DagRLEDecode(Data, Hdr.ImageSize, Size, Bits);
          end
          else
          begin
            // read compressed data Arena
            // compress algorhitm is unknown to me now
            // if Unk = 264 then after header is word size of original data
            // if Unk = 260 no size after head

            NewImage(Image, Width, Height, ifIndex8);
          end;

        // palette creation
        PalA := FPalette;
        if (Input.Size = Input.Position + 768) then
        begin
          ReadBuffer(Pal, SizeOf(Pal));
          PalA := RGBPalToARGBPal(Pal, $FF);
        end;

        PalSize := 256;
        GetMem(Palette, 256 * SizeOf(TARGBRec));
        Move(PalA, Palette^, SizeOf(PalA));
        PalSwapChannels(Palette, 256);

      except
        RaiseError(ErrString(Self, 'DecodeStream',
          'Error while loading Img data! Image can be corrupted.'),
            EDaggerfall);
        Input.Position := OldPos;
      end;
    finally

    end;
  Result := true;
end;


{*************************************************************************}
{                         TCifModule Methods                              }
{*************************************************************************}


constructor TCifModule.Create;
begin
  inherited Create;
  FInfo := CifModuleInfo;
end;

function TCifModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TCifModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  Hdr: TImgHeader;
  I: Long;
  OldPos: Int64;
  Clone: TImageData;

begin
  Result := false;
  Assert(Output <> nil, ErrString(Self, 'EncodeStream',
    'Output stream is nil!'));

  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  OldPos := OutPut.Position;
  PrepareImage(Image, Clone);

  with OutPut, Clone do
  try
    try

      for I := 0 to LevelCount - 1 do
        with Clone.Levels[I] do
        begin

          FillChar(Hdr, SizeOf(Hdr), 0);
          Hdr.Width := Width;
          Hdr.Height := Height;
          Hdr.ImageSize := Width * Height;

          WriteBuffer(Hdr, SizeOf(Hdr));
          WriteBuffer(Bits^, Hdr.ImageSize);
        end;

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving image data! Image can be corrupted.'), EDaggerfall);
      Output.Position := OldPos;
    end;
  finally
    FreeImageSafe(Clone);
  end;
  Result := true;
end;

function TCifModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TCifModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  Hdr: TImgHeader;
  Grp: TCifGroup;
  HasHdr: boolean;
  IsRLE, IsWeapon, ISW9, IsStd, IsFirst: boolean;
  Data: Pointer;
  OldPos: Long;
  OldSPos: Int64;
  W, H, ALev, I, S: Long;
begin
  Result := false;
  FreeImageSafe(Image);

  OldSPos := Input.Position;
  with Input, Image do
  try
    try
      // palette settings
      BitCount := 8;
      //PixelFormat := Images.PixelFormat(0, 8, 8, 8);
      ImageFormat := ifIndex8;
      PalSize := 256;
      GetMem(Palette, 256 * SizeOf(TARGBRec));
      Move(FPalette, Palette^, SizeOf(FPalette));
      PalSwapChannels(Palette, 256);

      W := 0;
      H := 0;

      // solve cif type
      HasHdr := true;
      IsRLE := false;
      IsWeapon := false;
      IsW9 := false;
      IsFirst := true;

      ReadBuffer(Hdr, SizeOf(Hdr));
      if Hdr.XOff = 40347 then
      begin
        // image has no header (FACES.CIF only)
        W := 64;
        H := 64;

        HasHdr := false;
      end;
      if Hdr.Unk = $15 then
      begin
        // this file is Weapon09.cif
        IsWeapon := true;
        IsW9 := true;
      end;
      if HasHdr then
        if Position + Hdr.ImageSize < Input.Size then
        begin
          Position := Position + Hdr.ImageSize;
          ReadBuffer(Grp, SizeOf(Grp));
          if Grp.Offsets[0] = 76 then
          // cif is regular weapon
            IsWeapon := true;
        end;

      IsStd := HasHdr and (not IsWeapon);
      Position := 0;

      while Position < Input.Size do
      begin
        // process standard cif file
        if IsStd then
        begin
          ReadBuffer(Hdr, SizeOf(Hdr));
          AddLevel(Image);
          ALev := Image.LevelCount - 1;
          Levels[ALev].Width := Hdr.Width;
          Levels[ALev].Height := Hdr.Height;
          Levels[ALev].Size := Hdr.Width * Hdr.Height;

          GetMem(Data, Hdr.ImageSize);
          ReadBuffer(Data^, Hdr.ImageSize);

          if Hdr.Unk <> 2 then
          begin
            // uncompressed
            GetMem(Levels[ALev].Bits, Levels[ALev].Size);
            Move(Data^, Levels[ALev].Bits^, Levels[ALev].Size);
          end
          else
            //rle compressed
            DagRLEDecode(Data, Hdr.ImageSize, Levels[ALev].Size,
              Levels[ALev].Bits);

          FreeMem(Data);
        end;

        // process file without header
        if not HasHdr then
        begin
          AddLevel(Image);
          ALev := Image.LevelCount - 1;
          Levels[ALev].Width := W;
          Levels[ALev].Height := H;
          Levels[ALev].Size := W * H;

          GetMem(Levels[ALev].Bits, Levels[ALev].Size);
          ReadBuffer(Levels[ALev].Bits^, Levels[ALev].Size);
        end;

        // process weapon file
        if IsWeapon then
          // first frame is std IMG file, next are not
          // but if IsW9 this first frame is missing
          if IsFirst and (not IsW9) then
          begin
            ReadBuffer(Hdr, SizeOf(Hdr));
            AddLevel(Image);
            ALev := Image.LevelCount - 1;
            Levels[ALev].Width := Hdr.Width;
            Levels[ALev].Height := Hdr.Height;
            Levels[ALev].Size := Hdr.ImageSize;

            GetMem(Levels[ALev].Bits, Levels[ALev].Size);
            ReadBuffer(Levels[ALev].Bits^, Levels[ALev].Size);

            IsFirst := false;
          end
          else
          begin
            OldPos := Position;
            // read next group
            ReadBuffer(Grp, SizeOf(Grp));
            // read images in group
            I := 0;
            while Grp.Offsets[I] <> 0 do
            begin
              S := Grp.Offsets[I + 1] - Grp.Offsets[I];
              if S < 0 then
                S := Grp.Offsets[31] - Grp.Offsets[I];
              Position := OldPos + Grp.Offsets[I];
              AddLevel(Image);
              ALev := Image.LevelCount - 1;
              Levels[ALev].Width := Grp.Width;
              Levels[ALev].Height := Grp.Height;
              Levels[ALev].Size := Grp.Width * Grp.Height;

              GetMem(Data, S);
              ReadBuffer(Data^, S);

              DagRLEDecode(Data, Grp.ImageSize, Levels[ALev].Size,
                Levels[ALev].Bits);
              FreeMem(Data);

              Inc(I);
            end;
            Position := OldPos + Grp.Offsets[31];
          end;
      end;

    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading Cif data! Image can be corrupted.'), EDaggerfall);
      Input.Position := OldSPos;
    end;
  finally

  end;
  Result := true;
end;


{*************************************************************************}
{                         TTextureModule Methods                          }
{*************************************************************************}


constructor TTextureModule.Create;
begin
  inherited Create;
  FInfo := TexModuleInfo;
  FCanEncode := false;
  FLastName := 'Unknown';
end;

function TTextureModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
begin
  Result := false;
  RaiseError(ErrString(Self, 'EncodeBuffer',
    'Daggerfall texture module does not support saving !'), EDaggerfall);
end;

function TTextureModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
begin
  Result := false;
  RaiseError(ErrString(Self, 'EncodeBuffer',
    'Daggerfall texture module does not support saving !'), EDaggerfall);
end;

function TTextureModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TTextureModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  Hdr: TTexHeader;
  List: POffsetList;
  IHdr: TTexImgHeader;
  I, HPos: Long;
  OldPos: Int64;

  procedure LoadUncomp;
  var
    I: Long;
  begin
    AddLevel(Image);
    with Image.Levels[Image.LevelCount - 1] do
    begin
      Width := IHdr.Width;
      Height := IHdr.Height;
      Size := Width * Height;
      GetMem(Bits, Size);
      for I := 0 to Height - 1 do
      begin
        Input.ReadBuffer(Pointer(Long(Bits) + I * Width)^, Width);
        Input.Position := Input.Position + 256 - Width;
      end;
    end;
  end;

  procedure LoadUnSubs;
  var
    SubOffs: packed array[0..63] of Long;
    I, Start, J, RPos: Long;
    Z, NZ: Byte;
    W, H: Word;

  begin

    // read subimages offset list
    Start := Input.Position;
    FillChar(SubOffs, SizeOf(SubOffs), 0);
    Input.ReadBuffer(SubOffs, IHdr.SubImages * 4);

    // process subimages
    for I := 0 to IHdr.SubImages - 1 do
    begin
      AddLevel(Image);
      with Image.Levels[Image.LevelCount - 1] do
      begin
        Width := IHdr.Width;
        Height := IHdr.Height;
        Size := Width * Height;
        GetMem(Bits, Size);

        Input.Position := Start + SubOffs[I];
        Input.ReadBuffer(W, 2);
        Input.ReadBuffer(H, 2);
        // read rows
        for J := 0 to H - 1 do
        begin
          RPos := 0;
          while RPos < W do
          begin
            Input.ReadBuffer(Z, 1);
            FillChar(PByte(Long(Bits) + J * W + RPos)^, Z, 0);
            RPos := RPos + Z;
            Input.ReadBuffer(NZ, 1);
            Input.ReadBuffer(PByte(Long(Bits) + J * W + RPos)^, NZ);
            RPos := RPos + NZ;
          end;
        end;
      end;
    end;
  end;

  procedure LoadRLESubs;

  type
    TRowOff = packed record
      Off: Word;
      Null: Byte;
      RLEIdx: Byte;
    end;

  var
    RowOffs: packed array[0..255] of TRowOff;
    I, J, RPos, NextOff: Long;
    Bytec, C, Ind: Byte;
  begin
    NextOff := Input.Position;

    // process subimages
    for I := 0 to IHdr.SubImages - 1 do
    begin
      FillChar(RowOffs, SizeOf(RowOffs), 0);
      Input.Position := NextOff;
      Input.ReadBuffer(RowOffs, IHdr.Height * SizeOf(TRowOff));
      NextOff := Input.Position;

      AddLevel(Image);
      with Image.Levels[Image.LevelCount - 1] do
      begin
        Width := IHdr.Width;
        Height := IHdr.Height;
        Size := Width * Height;
        GetMem(Bits, Size);
        // read rows
        for J := 0 to Height - 1 do
        begin
          Input.Position := HPos + RowOffs[J].Off;
          if RowOffs[J].RLEIdx = 128 then
          begin
            Input.ReadBuffer(C, 2);
            RPos := 0;
            while RPos < Width do
            begin
              Input.ReadBuffer(Bytec, 1);
              Input.ReadBuffer(Ind, 1);
              if Ind = 0 then
              begin
                Input.ReadBuffer(PByte(Long(Bits) + J * Width + RPos)^,
                  Bytec);
                RPos := RPos + Bytec;
              end
              else
              begin
                Input.ReadBuffer(C, 1);
                FillChar(PByte(Long(Bits) + J * Width + RPos)^, 256 - Bytec,
                  C);
                RPos := RPos + 256 - Bytec;
              end;
            end;
          end
          else
            Input.ReadBuffer(PByte(Long(Bits) + J * Width)^, Width);
        end;
      end;
    end;
  end;

begin
  Result := false;
  FreeImageSafe(Image);

  OldPos := Input.Position;

  with Input, Image do
  try
    try
      BitCount := 8;
      //PixelFormat := Images.PixelFormat(0, 8, 8, 8);
      ImageFormat := ifIndex8;
      PalSize := 256;
      GetMem(Palette, 256 * SizeOf(TARGBRec));
      Move(FPalette, Palette^, SizeOf(FPalette));
      PalSwapChannels(Palette, 256);

      // handle texture.001 and texture.000 which contain color info
      if Input.Size = 2586 then
        NewImage(Image, 8, 8, ifIndex8)
      else
        if Input.Size = 46 then
          NewImage(Image, 8, 8, ifIndex8)
        else
          if Input.Size = 266 then
            NewImage(Image, 8, 8, ifIndex8)
          else
          begin
            ReadBuffer(Hdr, SizeOf(Hdr));
            FLastName := RepairName(Hdr.TexName);
            GetMem(List, Hdr.ImgCount * SizeOf(TOffset));
            // load offsets
            for I := 0 to Hdr.ImgCount - 1 do
              ReadBuffer(List[I], SizeOf(TOffset));

            // process offsets
            for I := 0 to Hdr.ImgCount - 1 do
            begin
              // jump at position of image header
              Position := List[I].HdrOffset;
              HPos := Position;
              ReadBuffer(IHdr, SizeOf(IHdr));

              begin
                Position := HPos + IHdr.ImageOff;

                if (IHdr.SubImages = 1) and (IHdr.Unk1 <> $1108) and (IHdr.Unk1
                  <> $0108) then
                  LoadUncomp;
                if (IHdr.SubImages = 1) and (IHdr.Unk1 = $1108) then
                  LoadRLESubs;
                if (IHdr.SubImages > 1) and (IHdr.Unk1 <> $0108) then
                  LoadUnSubs;
                if (IHdr.Unk1 = $0108) then
                  LoadRLESubs;
              end;
            end;
            FreeMem(List);
          end;
    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading Texture data! Image can be corrupted.'),
          EDaggerfall);
      Input.Position := OldPos;
    end;
  finally

  end;
  Result := true;
end;


initialization

  ImgModuleInfo:=GetDefaultIntModuleInfo;
  ImgModuleInfo.Name:=sImgName;
  ImgModuleInfo.Description:=sImgDesc;
  ImgModuleInfo.IDCode:=sImgCode;
  RegisterIntModule(@ImgModuleInfo,TImgModule);

  CifModuleInfo:=GetDefaultIntModuleInfo;
  CifModuleInfo.Name:=sCifName;
  CifModuleInfo.Description:=sCifDesc;
  CifModuleInfo.IDCode:=sCifCode;
  RegisterIntModule(@CifModuleInfo,TCifModule);

  TexModuleInfo:=GetDefaultIntModuleInfo;
  TexModuleInfo.Name:=sTexName;
  TexModuleInfo.Description:=sTexDesc;
  TexModuleInfo.IDCode:=sTexCode;
  RegisterIntModule(@TexModuleInfo,TTextureModule);

end.



