{*************************************************************************}
{*   JpgModule                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images\Jpg\JpgModule.pas                                    *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 9th March 2003  (09.03.2003)                               }
{*************************************************************************}


{This unit contains module for loading/saving Jpeg images.}


unit JpgModule;


interface


uses
  Sys, Storage, Images, Objects, SysUtils, 
 {Uses modified JPEG library shipped with Delphi}
  jpeglib;


type

  {Image encoder for loading/saving Jpg file format.
  Supports: all types loading/saving.}
  TJpgModule = class(TImageModule)
  protected

    FGreyScale: boolean;
    FProgressive: boolean;
    FQuality: Long;

    function BuildPalette(const cinfo: jpeg_decompress_struct): PPalArray;
    procedure InitDecompressor(Stream: TStream; var jc: TJPEGContext);
    procedure ReleaseContext(var jc: TJPEGContext);

    {Here convert from image formats not supported by encoder }
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      override;

  public

    constructor Create; override;
    destructor Destroy; override;

    {This function converts ImageData structure to Jpg.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This function converts Jpg data to ImageData structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func takes valid Jpg image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream): boolean;
      override;

    {This func converts Jpg in Input stream read from image file to Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;

    {This function tests if data are valid Jpg image.}
    function TestFormat(Data: Pointer; Size: Long): boolean; override;

    {This function tests if stream contains valid Jpg.}
    function TestFormat(Input: TStream): boolean; override;


    {If set module will encode images as 8 bit greyscale images.}
    property GreyScale: boolean read FGreyScale write FGreyScale;

    {If set module will use progressive encoding when writting jpegs.}
    property Progressive: boolean read FProgressive write FProgressive;

    {Defines jpeg compression quality  1..100(best) .}
    property Quality: Long read FQuality write FQuality;


  end;


const

  sJpgName = 'Jpg Module';
  sJpgDesc =
    'Provides Joint Photographic Experts Group (*.jpg) image format encoding/decoding';
  sJpgCode: TIDCode = 'IMG-JPEG';

  { JPG file identifier}
  JPGMagic:array[0..3]of Char=#$FF#$D8#$FF#$E0;


var

  JpgModuleInfo:TModuleInfo;

implementation


procedure JpegError(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure EmitMessage(cinfo: j_common_ptr; msg_level: Integer);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure OutputMessage(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure FormatMessage(cinfo: j_common_ptr; buffer: PChar);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
end;

procedure ResetErrorMgr(cinfo: j_common_ptr);
{$IFDEF LINUX} cdecl; export;
{$ENDIF}
begin
  cinfo^.err^.num_warnings := 0;
  cinfo^.err^.msg_code := 0;
end;

const
  jpeg_std_error: jpeg_error_mgr = (
    error_exit: JpegError;
    emit_message: EmitMessage;
    output_message: OutputMessage;
    format_message: FormatMessage;
    reset_error_mgr: ResetErrorMgr);


{*************************************************************************}
{                         TJpgModule Methods                              }
{*************************************************************************}


constructor TJpgModule.Create;
begin
  inherited Create;
  FInfo := JpgModuleInfo;
  FGreyScale := false;
  FProgressive := false;
  FQuality := 90;
end;

destructor TJpgModule.Destroy;
begin

  inherited Destroy;
end;

procedure TJpgModule.InitDecompressor(Stream: TStream; var jc: TJPEGContext);
begin
  FillChar(jc, sizeof(jc), 0);
  jc.err := jpeg_std_error;
  jc.common.err := @jc.err;

  jpeg_CreateDecompress(jc.d, JPEG_LIB_VERSION, sizeof(jc.d));
  try
{$IFDEF MSWINDOWS}
    jpeg_stdio_src(jc.d, Stream);
{$ELSE}
    JpegStdioSrc(jc.d, Stream);
{$ENDIF}
    jpeg_read_header(jc.d, TRUE);

    jc.d.scale_num := 1;
    jc.d.scale_denom := 1;
    jc.d.do_block_smoothing := true;

    if jc.d.out_color_space = JCS_GRAYSCALE then
    begin
      jc.d.quantize_colors := True;
      jc.d.desired_number_of_colors := 256;
    end;

    jc.FinalDCT := jc.d.dct_method;
    jc.FinalTwoPassQuant := jc.d.two_pass_quantize;
    jc.FinalDitherMode := jc.d.dither_mode;
    if jpeg_has_multiple_scans(jc.d) then
    begin
      jc.d.enable_2pass_quant := jc.d.two_pass_quantize;
      jc.d.dct_method := JDCT_IFAST;
      jc.d.two_pass_quantize := False;
      jc.d.dither_mode := JDITHER_ORDERED;
      jc.d.buffered_image := True;
    end;
  except
    ReleaseContext(jc);
    RaiseError(ErrString(Self, 'InitDecompressor',
      'Error when initializing JPEG decompressor!'),EImage);
  end;
end;

procedure TJpgModule.ReleaseContext(var jc: TJPEGContext);
begin
  if jc.common.err = nil then
    Exit;
  jpeg_destroy(jc.common);
  jc.common.err := nil;
end;

function TJpgModule.BuildPalette(const cinfo: jpeg_decompress_struct):
  PPalArray;
var
  I: Long;
  C: Byte;
  Pal: PPalArray;
begin
  GetMem(Pal, cinfo.actual_number_of_colors * SizeOf(TARGBRec));
  if cinfo.out_color_space = JCS_GRAYSCALE then
    for I := 0 to cinfo.actual_number_of_colors - 1 do
    begin
      C := cinfo.colormap^[0]^[I];
      Pal[I].A := $FF;
      Pal[I].R := C;
      Pal[I].G := C;
      Pal[I].B := C;
    end
  else
    for I := 0 to cinfo.actual_number_of_colors - 1 do
    begin
      Pal[I].A := $AA;
      Pal[I].R := cinfo.colormap^[2]^[I];
      Pal[I].G := cinfo.colormap^[1]^[I];
      Pal[I].B := cinfo.colormap^[0]^[I];
    end;
  Result := Pal;
end;

procedure TJpgModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone do
  begin
    // JPG supports 8grey/24bit images only but input must
    //   be always 24bit
    FGreyScale := false;
    case BitCount of
      8:
        begin
          if ImageFormat = ifIntensity8 then
            // image will be saved as 8 bit greyscale
            FGreyScale := true;
          Convert8To24(Clone);
        end;
      16: Convert16To24(Clone);
      32: Convert32To24(Clone);
    end;
  end;
end;

function TJpgModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TJpgModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  OldPos: Int64;
  LinesPerCall, LinesRead: Long;
  Dest, CurPixel: PChar;
  Swap: Char;
  PtrInc: Long;
  jc: TJPEGContext;

begin
  Result := false;
  FreeImageSafe(Image);
  Assert(TestFormat(Input), ErrString(Self, 'DecodeStream',
    'Buffer image data not in Jpg format or corrupted!'));

  OldPos := Input.Position;

  AddLevel(Image);
  InitDecompressor(Input, jc);
  with Image, Image.Levels[0] do
  try
    try

      if (jc.d.out_color_space = JCS_GRAYSCALE) then
        BitCount := 8
      else
        BitCount := 24;

      jpeg_start_decompress(jc.d);

      //PixelFormat:=Images.PixelFormat(0,8,8,8);
      ImageFormat := Iff(BitCount = 24, ifR8G8B8, ifIntensity8);
      Width := jc.d.output_width;
      Height := jc.d.output_height;
      Size := Width * Height * (BitCount shr 3);
      GetMem(Bits, Size);
      Dest := Bits;
      PtrInc := Width * (BitCount shr 3);
      LinesPerCall := jc.d.rec_outbuf_height;

      // support for progressive images
      if jc.d.buffered_image then
      begin
        while jpeg_consume_input(jc.d) <> JPEG_REACHED_EOI do
        begin
          jpeg_start_output(jc.d, jc.d.input_scan_number);
          Dest := Bits;
          while (jc.d.output_scanline < jc.d.output_height) do
          begin
            LinesRead := jpeg_read_scanlines(jc.d, @Dest, LinesPerCall);
{$IFDEF LINUX}
            if BitCount = 24 then
            begin
              CurPixel := DestScanLine;
              while (CurPixel - DestScanLine) < (Width * 3) do
              begin
                Swap := CurPixel[0];
                CurPixel[0] := CurPixel[2];
                CurPixel[2] := Swap;
                Inc(CurPixel, 3);
              end;
            end;
{$ENDIF}
            Inc(Long(Dest), PtrInc * LinesRead);
          end;
          jpeg_finish_output(jc.d);
        end;
        // reset options for final pass at requested quality
        jc.d.dct_method := jc.FinalDCT;
        jc.d.dither_mode := jc.FinalDitherMode;
        if jc.FinalTwoPassQuant then
        begin
          jc.d.two_pass_quantize := True;
          jc.d.colormap := nil;
        end;
        jpeg_start_output(jc.d, jc.d.input_scan_number);
        Dest := Bits;
      end;

      // build final color palette
      //   Disabled now. GrayScale images are now in ifIntensity8 format
      {if (not jc.d.buffered_image or jc.FinalTwoPassQuant) and
        (jc.d.colormap <> nil) then
      begin
        PalSize := jc.d.actual_number_of_colors;
        Palette := BuildPalette(jc.d);
        Dest := Bits;
      end;}

        // final image pass for progressive, first and only pass for baseline
      while (jc.d.output_scanline < jc.d.output_height) do
      begin
        LinesRead := jpeg_read_scanlines(jc.d, @Dest, LinesPerCall);
{$IFDEF LINUX}
        if BitCount = 24 then
        begin
          CurPixel := Dest;
          while (CurPixel - DestScanLine) < (Width * 3) do
          begin
            Swap := CurPixel[0];
            CurPixel[0] := CurPixel[2];
            CurPixel[2] := Swap;
            Inc(CurPixel, 3);
          end;
        end;
{$ENDIF}
        Inc(Long(Dest), PtrInc * LinesRead);
      end;

      if jc.d.buffered_image then
        jpeg_finish_output(jc.d);
      jpeg_finish_decompress(jc.d);

    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading Jpg data! Image can be corrupted.'),EImage);
      Input.Position := OldPos;
    end;
  finally
    ReleaseContext(jc);
  end;
  Result := true;
end;

function TJpgModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TJpgModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  LinesWritten, LinesPerCall: Long;
  Src: Pointer;
  PtrInc: Long;
  jc: TJPEGContext;
  Clone: TImageData;
  OldPos: Int64;
begin
  Result := false;
  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  OldPos := Output.Position;

  PrepareImage(Image, Clone);

  with Clone, Clone.Levels[0] do
  try
    try
      FillChar(jc, sizeof(jc), 0);
      jc.err := jpeg_std_error;
      jc.common.err := @jc.err;

      jpeg_CreateCompress(jc.c, JPEG_LIB_VERSION, sizeof(jc.c));

{$IFNDEF LINUX}
      jpeg_stdio_dest(jc.c, OutPut);
{$ELSE}
      JPEGStdioDest(jc.c, OutPut);
{$ENDIF}

      jc.c.image_width := Width;
      jc.c.image_height := Height;
      jc.c.input_components := 3;   // JPEG requires 24bit RGB input
      jc.c.in_color_space := JCS_RGB;

      jpeg_set_defaults(jc.c);
      jpeg_set_quality(jc.c, FQuality, True);

      if FGreyScale then
        jpeg_set_colorspace(jc.c, JCS_GRAYSCALE);

      if FProgressive then
        jpeg_simple_progression(jc.c);

      Src := Bits;
      PtrInc := Width * (BitCount shr 3);
      LinesPerCall := 1;

      jpeg_start_compress(jc.c, true);

      while (jc.c.next_scanline < jc.c.image_height) do
      begin
        LinesWritten := jpeg_write_scanlines(jc.c, @Src,
          LinesPerCall);
        Inc(Long(Src), PtrInc * LinesWritten);
      end;

      jpeg_finish_compress(jc.c);

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving Jpg data! Image can be corrupted.'),EImage);
      OutPut.Position := OldPos;
    end;
  finally
    ReleaseContext(jc);
    FreeImageSafe(Clone);
  end;
  Result := true;
end;

function TJpgModule.TestFormat(Input: TStream): boolean;
var
  OldPos: Int64;
  C: array[0..3]of Char;
begin
  Result := false;
  if Input = nil then
    Exit;

  OldPos := Input.Position;
  Input.ReadBuffer(C, SizeOf(C));
  Input.Position := OldPos;

  if CompareMem(@C, @JPGMagic, SizeOf(C)) then
    Result := true;
end;

function TJpgModule.TestFormat(Data: Pointer; Size: Long): boolean;
begin
  Result := false;
  if (Data = nil) or (Size = 0) then
    Exit;

  if CompareMem(PULong(Data), @JPGMagic, SizeOf(JPGMagic)) then
    Result := true;
end;


initialization

  JpgModuleInfo:=GetDefaultIntModuleInfo;
  JpgModuleInfo.Name:=sJpgName;
  JpgModuleInfo.Description:=sJpgDesc;
  JpgModuleInfo.IDCode:=sJpgCode;
  RegisterIntModule(@JpgModuleInfo,TJpgModule);

end.

