{*************************************************************************}
{*   TgaModule                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Graphic\Images\Tga\TgaModule.pas                                    *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 22th March 2003  (22.03.2003)                              }
{*************************************************************************}


{This unit contains module for loading/saving Targa files.}


unit TgaModule;


interface


uses
 Sys,Objects,Images,SysUtils,Maths,Storage;


type

  {Image encoder for loading/saving targa file format.
  Supports: 1,8,16,24,32 bit bitmaps loading/saving.
  Compression: None, RLE }
  TTgaModule = class(TImageModule)
  protected

    FUseRLE: boolean;

    function GetUseRLE: boolean;

    procedure SetUseRLE(Value: boolean);

    {Here convert from image formats not supported by encoder }
    procedure PrepareImage(const Image: TImageData; out Clone: TImageData);
      override;


  public

    constructor Create; override;
    destructor Destroy; override;

    {This function converts ImageData structure to targa.}
    function EncodeBuffer(const Image: TImageData; out Data: Pointer; out Size:
      Long): boolean; override;

    {This func takes valid targa image data and converts it to specified image
    file format ready to be written into file.}
    function EncodeStream(const Image: TImageData; Output: TStream):
      boolean; override;


    {This function converts targa data to ImageData structure.}
    function DecodeBuffer(Data: Pointer; Size: Long; var Image: TImageData):
      boolean; override;

    {This func converts targa in Input stream read from image file to  Image Data
    structure.}
    function DecodeStream(Input: TStream; var Image: TImageData): boolean;
      override;


    {This function tests if data are valid targa image.}
    function TestFormat(Data: Pointer; Size: Long): boolean; override;

    {This function tests if stream contains valid targa.}
    function TestFormat(Input: TStream): boolean; override;


  published

     {Indicates that module uses RLE compression when saving targas.}
    property UseRLE: boolean read GetUseRLE write SetUSeRLE;

  end;


  {Targa file header}
  TTGAHeader = packed record
    IDLength: Byte;
    ColorMapType: Byte;
    ImageType: Byte;
    ColorMapOff: Word;
    ColorMapLength: Word;
    ColorEntrySize: Byte;
    XOrg: Small;
    YOrg: Small;
    Width: Small;
    Height: Small;
    PixelSize: Byte;
    Desc: Byte;
  end;


  {Footer at the end of TGA file}
  TTGAFooter = packed record
    ExtOff: ULong; // Extension Area Offset
    DevDirOff: ULong; // Developer Directory Offset
    Signature: array[0..15] of Char; // TRUEVISION-XFILE
    Reserved: Byte; // ASCII period '.'
    NullChar: Byte; // 0
  end;


const

  sTGASignature = 'TRUEVISION-XFILE';

  sTgaName = 'Tga Module';
  sTgaDesc = 'Provides Truevision targa (*.tga) image format encoding/decoding';
  sTgaCode: TIDCode = 'IMG-TGA';


var

  TgaModuleInfo:TModuleInfo;


implementation


{*************************************************************************}
{                         TTgaModule Methods                              }
{*************************************************************************}


constructor TTgaModule.Create;
begin
  inherited Create;
  FInfo := TgaModuleInfo;

  FUseRLE := true;
end;

destructor TTgaModule.Destroy;
begin

  inherited Destroy;
end;

function TTgaModule.GetUseRLE: boolean;
begin
  Result := FUseRLE;
end;

procedure TTgaModule.SetUseRLE(Value: boolean);
begin
  FUseRLE := Value;
end;

procedure TTgaModule.PrepareImage(const Image: TImageData; out Clone:
  TImageData);
begin
  inherited PrepareImage(Image, Clone);
  with Clone do
  begin
   case ImageFormat of
   Low(if8bitARGBPF) .. High(if8bitARGBPF):
      if ImageFormatToPF(ImageFormat).ABitCount > 0 then
       Convert8To32(Clone)
      else
       Convert8To24(Clone);
   end;
  end;
end;

function TTgaModule.EncodeBuffer(const Image: TImageData;
  out Data: Pointer; out Size: Long): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  try
    // calls EncodeStream function rather that try to parse raw pointer
    Result := EncodeStream(Image, MS);
  finally
    Size := MS.Size;
    GetMem(Data, Size);
    // this is important line
    MS.Position := 0;
    MS.ReadBuffer(Data^, Size);
    MS.Free;
  end;
end;

function TTgaModule.EncodeStream(const Image: TImageData;
  Output: TStream): boolean;
var
  StartPos: Int64;
  Hdr: TTGAHeader;
  I: Long;
  Pal: PPalRGBArray;
  Clone:TImageData;

  procedure SaveRGB;
  begin
    OutPut.WriteBuffer(Clone.Levels[0].Bits^, Clone.Levels[0].Size);
  end;

  procedure SaveRLE;
  var
    APix, Pix: PULong;
    D, S, Cnt, CPixel: Long;
    B, Rle: Byte;

    function CountSame: Long;
    var
      I: Long;
    begin
      I := CPixel;
      Result := 1;
      Move(Pointer(Long(Clone.Levels[0].Bits) + I * B)^, Pix^, B);
      while I < Cnt do
      begin
        I := I + 1;
        Move(Pointer(Long(Clone.Levels[0].Bits) + I * B)^, APix^, B);
        if (CompareMem(APix, Pix, B))and
        (Result<127) then
          Result := Result + 1
        else
          Break;
      end;
    end;

    function CountDiff: Long;
    var
      I: Long;
    begin
      I := CPixel;
      Result := 0;
      while I < Cnt do
      begin
        Move(Pointer(Long(Clone.Levels[0].Bits) + I * B)^, Pix^, B);
        I := I + 1;
        Move(Pointer(Long(Clone.Levels[0].Bits) + I * B)^, APix^, B);
        if (not CompareMem(APix, Pix, B))and
         (Result<127)then
          Result := Result + 1
        else
          Break;
      end;
    end;

  begin
    with Output, Clone, Clone.Levels[0] do
    begin
      B := BitCount shr 3;
      Cnt := Width * Height;
      CPixel := 0;
      GetMem(APix, 4);
      GetMem(Pix, 4);
      CPixel := 0;
      while CPixel < Cnt do
      begin
        D := CountDiff;
        S := CountSame;
        // these are now handled in CountDiff and CountSame - faster
        //D := Iff(D > 127, 127, D);
        //S := Iff(S > 127, 127, S);
        if D > 0 then
        begin
          Rle := D - 1;
          WriteBuffer(Rle, 1);
          WriteBuffer(Pointer(Long(Bits) + CPixel * B)^, D * B);
          CPixel := CPixel + D;
        end;
        if S > 1 then
        begin
          Rle := S + 127;
          WriteBuffer(Rle, 1);
          WriteBuffer(Pointer(Long(Bits) + CPixel * B)^, B);
          CPixel := CPixel + S;
        end;
      end;
      FreeMemNil(APix);
      FreeMemNil(Pix);
    end;
  end;

begin
  Result := false;
  Assert(Output <> nil, ErrString(Self, 'EncodeStream',
    'Output stream is nil!'));

  Assert(TestImage(Image), ErrString(Self, 'EncodeStream',
    'Image data information is invalid! Image parameter: ' + #10#13 +
    ImageToStr(Image)));

  StartPos:=OutPut.Position;

  PrepareImage(Image,Clone);

  with OutPut, Clone, Clone.Levels[0] do
  try
    try
      FillChar(Hdr, SizeOf(Hdr), 0);
      Hdr.IDLength := 0;
      Hdr.ColorMapType := Sys.Iff(PalSize > 0, 1, 0);
      Hdr.Width := Width;
      Hdr.Height := Height;
      Hdr.PixelSize := BitCount;
      Hdr.ColorMapLength := PalSize;
      Hdr.ColorEntrySize := Sys.Iff(PalSize > 0, 24, 0);
      Hdr.ColorMapOff := 0;
      // this indicates that targa is stored in top-left format
      // as our images -> no flipping is needed.
      Hdr.Desc := 32;

      case BitCount of
      8: case ImageFormat of
       ifIndex8:Hdr.ImageType := Sys.Iff(FUseRLE, 9, 1);
       ifIntensity8:Hdr.ImageType := Sys.Iff(FUseRLE, 11, 3);
      end;
      else
        Hdr.ImageType := Sys.Iff(FUseRLE, 10, 2);
      end;

      WriteBuffer(Hdr, SizeOf(Hdr));

      // writting palette
      if PalSize > 0 then
      begin
        GetMem(Pal, PalSize * SizeOf(TRGBRec));
        for I := 0 to PalSize - 1 do
          Pal[I] := RGBRec(Palette[I].R, Palette[I].G, Palette[I].B);
        WriteBuffer(Pal^, PalSize * SizeOf(TRGBRec));
        FreeMemNil(Pal);
      end;

      case Hdr.ImageType of
        1, 2,3:
            //save uncompressed mode images
          SaveRGB;
        9, 10,11:
            //save rle compressed mode images
          SaveRLE;
      end;

    except
      RaiseError(ErrString(Self, 'EncodeStream',
        'Error while saving image data! Image can be corrupted.'),EImage);

      Position := StartPos;
    end;
  finally
    FreeImageSafe(Clone);
    // all OK
  end;
end;

function TTgaModule.DecodeBuffer(Data: Pointer; Size: Long;
  var Image: TImageData): boolean;
var
  MS: TMemoryStream;

begin
  MS := TMemoryStream.Create;
  MS.WriteBuffer(Data^, Size);
  // this is important line
  MS.Position := 0;
  try
    // calls DecodeStream function rather that try to parse raw pointer
    Result := DecodeStream(MS, Image);
  finally
    MS.Free;
  end;
end;

function TTgaModule.DecodeStream(Input: TStream;
  var Image: TImageData): boolean;
var
  StartPos: Int64;
  Hdr: TTGAHeader;
  Foo: TTGAFooter;
  I, PSize, FSize: Long;
  Pal: Pointer;
  W: Word;

  procedure LoadRGB;
  begin
    Input.ReadBuffer(Image.Levels[0].Bits^, Image.Levels[0].Size);
  end;

  procedure LoadRLE;
  var
    I, CPixel, Cnt: Long;
    B, Rle: Byte;
    Pix: PULong;

  begin
    with Input, Image, Image.Levels[0] do
    begin
      Cnt := Width * Height;
      B := BitCount shr 3;
      GetMem(Pix, B);
      CPixel := 1;
      while CPixel <= Cnt do
      begin
        Rle := 0;
        ReadBuffer(Rle, 1);
        if Rle < 128 then
        begin
          // uncompressed pixel
          Rle := Rle + 1;
          for I := 0 to Rle - 1 do
          begin
            // read pixel
            ReadBuffer(Pix^, B);
            // write pixel

            // !! these lines are much faster than using Move procedure
            //!! Move(Pix^,Pointer(Long(Bits) + Pad+(CPixel-1) * B)^,B);
            PByte(Long(Bits) + (CPixel - 1) * B)^ := PByte(Long(Pix) +
              0)^;
            if B > 1 then
              PByte(Long(Bits) + (CPixel - 1) * B + 1)^ := PByte(Long(Pix)
                + 1)^;
            if B > 2 then
              PByte(Long(Bits) + (CPixel - 1) * B + 2)^ := PByte(Long(Pix)
                + 2)^;
            if B > 3 then
              PByte(Long(Bits) + (CPixel - 1) * B + 3)^ := PByte(Long(Pix)
                + 3)^;

            CPixel := CPixel + 1;
          end;
        end
        else
        begin
          // compressed pixels
          Rle := Rle - 127;
          // read pixel
          ReadBuffer(Pix^, B);
          for I := 0 to Rle - 1 do
          begin
            // write pixel
            PByte(Long(Bits) + (CPixel - 1) * B)^ := PByte(Long(Pix) +
              0)^;
            if B > 1 then
              PByte(Long(Bits) + (CPixel - 1) * B + 1)^ := PByte(Long(Pix)
                + 1)^;
            if B > 2 then
              PByte(Long(Bits) + (CPixel - 1) * B + 2)^ := PByte(Long(Pix)
                + 2)^;
            if B > 3 then
              PByte(Long(Bits) + (CPixel - 1) * B + 3)^ := PByte(Long(Pix) +
                3)^;

            CPixel := CPixel + 1;
          end;
        end;
      end;
      FreeMemNil(Pix);
    end;
  end;

begin
  Result := false;
  FreeImageSafe(Image);
  Assert(TestFormat(Input), ErrString(Self, 'DecodeStream',
    'Stream image data not in Targa format (may be types 11, 32, 33) or corrupted!'));

  StartPos := Input.Position;
  FSize := SizeOf(Hdr);

  // !!! Level 0 must be created before "with Image.Levels[0]" is used"  !!!
  AddLevel(Image);

  with Input, Image, Image.Levels[0] do
  try
    try
      ReadBuffer(Hdr, SizeOf(Hdr));

      Width := Hdr.Width;
      Height := Hdr.Height;
      BitCount := Hdr.PixelSize;
      Size := Width * Height * (BitCount shr 3);

      // skip image ID info
      Position := Position + Hdr.IDLength;
      Inc(FSize, Hdr.IDLength);

      if Hdr.ColorMapType = 1 then
      begin
        // read palette
        PSize := Hdr.ColorMapLength * (Hdr.ColorEntrySize shr 3);
        Inc(FSize, PSize);
        GetMem(Pal, PSize);
        ReadBuffer(Pal^, PSize);

        // process palette
        PalSize := Hdr.ColorMapLength;
        GetMem(Palette, PalSize * SizeOf(TARGBRec));

        for I := 0 to PalSize - 1 do
          case Hdr.ColorEntrySize of
            24: Palette[I] := ARGBRec($FF, PByte(Long(Pal) + I * 3 + 2)^,
                PByte(Long(Pal) + I * 3 + 1)^, PByte(Long(Pal) + I * 3)^);
            // I've never seen tga with these palettes
            15, 16, 32: ;
          end;
        FreeMemNil(Pal);
      end;

      // reading image data
      GetMem(Bits, Size);
      case Hdr.ImageType of
        0, 1, 2, 3:
          //load uncompressed mode images
          LoadRGB;
        9, 10, 11:
          //load rle compressed mode images
          LoadRLE;
      end;
      Inc(FSize, Position - StartPos - FSize);

      case BitCount of
        8: ImageFormat := ifIndex8;
        16: ImageFormat := ifX1R5G5B5;
        24: ImageFormat := ifR8G8B8;
        32: ImageFormat := ifA8R8G8B8;
      end;

      // set format for Black-White targas
      if Hdr.ImageType in [3, 11] then
       case BitCount of
        8: ImageFormat := ifIntensity8;
        16: ImageFormat := ifIntensity16;
      end;

      // we must find true end of file and set Input.Position to it
      // paint programs appends extra info at the end of TGAs
      if Position + 2 < Input.Size then
      begin
        ReadBuffer(W, 2);
          // 495 = size of Extension Area
        if W = 495 then
        begin
          Inc(FSize, 495);
          Position := StartPos + FSize;
        end
        else
          Position := Position - 2;

        if Position + SizeOf(Foo) <= Input.Size then
        begin
          ReadBuffer(Foo, SizeOf(Foo));
          if Foo.Signature = sTGASignature then
          begin
            Inc(FSize, SizeOf(Foo));
            Position := StartPos + FSize;
          end
          else
            Position := Position - SizeOf(Foo);
        end;
      end;

        {DONE -o Galfar: flip image according to TTGAHeader.Desc}
      if Hdr.Desc < 31 then
        FlipYImage(Image);

      Position := StartPos + FSize;
    except
      RaiseError(ErrString(Self, 'DecodeStream',
        'Error while loading targa data! Image can be corrupted.'),EImage);
      Position := StartPos;
    end;
  finally

  end;
  Result := true;
end;

function TTgaModule.TestFormat(Input: TStream): boolean;
var
  Hdr: TTGAHeader;
  OldPos: Int64;
begin
  Result := false;
  if Input = nil then
    Exit;
  if Input.Size < SizeOf(Hdr) then
    Exit;

  OldPos := Input.Position;
  Input.ReadBuffer(Hdr, SizeOf(Hdr));
  Input.Position := OldPos;

  if not (Hdr.ImageType in [0, 1, 2, 3, 9, 10, 11]) then
    Exit;
  if not (Hdr.PixelSize in [1, 4, 8, 15, 16, 24, 32]) then
    Exit;

  Result := true;
end;

function TTgaModule.TestFormat(Data: Pointer; Size: Long): boolean;
var
  Hdr: TTGAHeader;
begin
  Result := false;
  if Data = nil then
    Exit;
  if Size < SizeOf(Hdr) then
    Exit;

  Move(Data^, Hdr, SizeOf(HDR));

  if not (Hdr.ImageType in [0, 1, 2, 3, 9, 10,11]) then
    Exit;
  if not (Hdr.PixelSize in [1, 4, 8, 15, 16, 24, 32]) then
    Exit;

  Result := true;
end;


initialization

  TgaModuleInfo:=GetDefaultIntModuleInfo;
  TgaModuleInfo.Name:=sTgaName;
  TgaModuleInfo.Description:=sTgaDesc;
  TgaModuleInfo.IDCode:=sTgaCode;
  RegisterIntModule(@TgaModuleInfo,TTgaModule);

end.
 