{*************************************************************************}
{*   Math                                                                *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Sys\Math.pas                                                        *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 16th February 2003  (16.02.2003)                           }
{*************************************************************************}


{This unit contains amount of mathematical functions.}


unit Maths;


interface


uses
  Sys;


const


  {Various mathematical constants}
  PI: Float = 3.1415926535897932384626433832795; // PI
  PIdiv2: Float = 1.5707963267948966192313216916398; // PI / 2
  PIdiv3: Float = 1.0471975511965977461542144610932; // PI / 3
  PIdiv4: Float = 0.78539816339744830961566084581988; // PI / 4
  PIdiv180:Float = 0.017453292519943295769236907684886; // PI / 180
  D180divPI:Float = 57.295779513082320876798154814105; // 180 / PI
  Sqrt2: Float = 1.4142135623730950488016887242097; // Sqrt(2)
  Sqrt3: Float = 1.7320508075688772935274463415059; // Sqrt(3)
  Sqrt5: Float = 2.2360679774997896964091736687313; // Sqrt(5)
  Sqrt10: Float = 3.1622776601683793319988935444327; // Sqrt(10)
  SqrtPI: Float = 1.7724538509055160272981674833411; // Sqrt(PI)
  Sqrt2PI: Float = 2.506628274631000502415765284811; // Sqrt(2 * PI)
  PI2: Float = 6.283185307179586476925286766559; // 2 * PI
  PI3: Float = 9.4247779607693797153879301498385; // 3 * PI
  Ln2: Float = 0.69314718055994530941723212145818; // Ln(2)
  Ln10: Float = 2.3025850929940456840179914546844; // Ln(10)
  LnPI: Float = 1.1447298858494001741434273513531; // Ln(PI)
  LogPI: Float = 0.4971498726941338543512682882909; // Log10(PI)
  LogE: Float = 0.43429448190325182765112891891661; // Log10(E)
  E: Float = 2.7182818284590452353602874713527; // Natural constant
  One:Float = 1.0000000000000000000000000000000;
  Intfinity:Float = 1/0;
  Nan:Float = 0/0;


{*************************************************************************}
{                         Base functions                                  }
{*************************************************************************}


{Returns Min value from array}
function MinFrom(const Data: array of Long): Long; overload;
function MinFrom(const Data: array of Float): Float; overload;

{Returns Max value from array}
function MaxFrom(const Data: array of Long): Long; overload;
function MaxFrom(const Data: array of Float): Float; overload;

{Returns true if value is in range (inclusive)}
function InRange(Value, Min, Max: Long): Boolean; overload;
function InRange(Value, Min, Max: Float): Boolean; overload;

{Returns sum of values}
function SumI(const Data: array of Long): Long;
function SumF(const Data: array of Float): Float;

{Returns sum of squares of values}
function SumSquaresI(const Data: array of Long): Long;
function SumSquaresF(const Data: array of Float): Float;

{Rounds float to specific digit}
function RoundTo(Value: Float; Digit: Short = -2): Float;

{Returns smallest integer >= Val}
function Ceil(Value: Float): Long;

{Returns largest integer <= Val}
function Floor(Value: Float): Long;

{Returns float division remainder}
function ModF(A, B: Float): Float;


{*************************************************************************}
{                Logarithmical/Exponential functions                      }
{*************************************************************************}


{Returns Log base 10 of X}
function Log10(X: Float): Float;

{Returns Log base 2 of X}
function Log2(X: Float): Float;

{Returns Log base N of X}
function LogN(Base, X: Float): Float;

{Raise base to long}
function IntPower(Base: Float; Exponent: Long): Float;

{Raise base to any power}
function Power(Base, Exponent: Float): Float;

{Raise ten to Exp}
function Power10(Exponent: Float): Float;

{Raise two to Exp}
function Power2(Exponent: Float): Float;

{Returns nearest greater power of 2. Mainly used when resizing
textures.}
function NearestPower2(Value:Long):Long;


{*************************************************************************}
{                      Random number functions                            }
{*************************************************************************}


{Returns long Min <= X <= Max}
function RandRange(Min, Max: Long): Long; overload;
function RandRange(Min, Max: Float): Float; overload;

{Returns value with random sign + / -}
function RandSign(Value: Long): Long; overload;
function RandSign(Value: Float): Float; overload;

{Returns random value from array}
function RandomFromI(const Values: array of Long): Long; overload;
function RandomFromF(const Values: array of Float): Float; overload;


{*************************************************************************}
{                      Trigonometric functions                            }
{*************************************************************************}


{Converts degrees to radians.}
function DegToRad(Degrees : Float) : Float;

{Converts radians to degrees.}
function RadToDeg(Radians: Float): Float;

{Returns sine and cosine of theta angle (in radians)}
procedure SinCos(Angle: Float; var Sin, Cos: Double); overload;
procedure SinCos(Angle: Float; var Sin, Cos: Single); overload;

{Use standard Delphi Sin/Cos functions - they are faster}
//function Sin
//function Cos

{Returns tangent of angle.}
function Tan(Angle : Float) : Float;

{Returns cotangent of angle.}
function CoTan(Angle: Float) : Float;

{Returns inverse cosine (in radians) of Value.}
function ArcSin(Value : Float) : Float;

{Returns inverse sine (in radians) of Value.}
function ArcCos(Value : Float): Float;

{Returns inverse tangent (in radians) of Value.}
function ArcTan(Value: Float): Float;

{Returns inverse cotangent (in radians) of Value.}
function ArcCoTan(Value: Float): Float;


implementation


{*************************************************************************}
{                         Base functions                                  }
{*************************************************************************}


function MinFrom(const Data: array of Long): Long;
var
  I: Long;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

function MinFrom(const Data: array of Float): Float;
var
  I: Long;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

function MaxFrom(const Data: array of Long): Long;
var
  I: Long;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;

function MaxFrom(const Data: array of Float): Float;
var
  I: Long;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;

function InRange(Value, Min, Max: Long): Boolean;
begin
  Result := (Value >= Min) and (Value <= Max);
end;

function InRange(Value, Min, Max: FLoat): Boolean;
begin
  Result := (Value >= Min) and (Value <= Max);
end;

function SumI(const Data: array of Long): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(Data) to High(Data) do
    Result := Result + Data[I];
end;

function SumF(const Data: array of Float): Float;
var
  I: Long;
begin
  Result := 0.0;
  for I := Low(Data) to High(Data) do
    Result := Result + Data[I];
end;

function SumSquaresI(const Data: array of Long): Long;
var
  I: Long;
begin
  Result := 0;
  for I := Low(Data) to High(Data) do
    Result := Result + Sqr(Data[I]);
end;


function SumSquaresF(const Data: array of Float): Float;
var
  I: Long;
begin
  Result := 0.0;
  for I := Low(Data) to High(Data) do
    Result := Result + Sqr(Data[I]);
end;

function RoundTo(Value: Float; Digit: Short): Float;
var
  LFactor: Double;
begin
  LFactor := IntPower(10, Digit);
  Result := Trunc((Value / LFactor) + 0.5) * LFactor;
end;

function Ceil(Value: Float): Long;
begin
  Result := Long(Trunc(Value));
  if Frac(Value) > 0 then
    Inc(Result);
end;

function Floor(Value: Float): Long;
begin
  Result := Long(Trunc(Value));
  if Frac(Value) < 0 then
    Dec(Result);
end;

function ModF(A, B: Float): Float;
var
  Z: Float;
begin
  Result := A / B;
  Z := Int(Result);
  if Result < 0.0 then
    Z := Z - 1.0;
  Result := A - Z * B;
end;


{*************************************************************************}
{                   Logarithmical/Exponential functions                   }
{*************************************************************************}


function Log10(X: Float): Float;
asm
    FLDLG2
    FLD     X
    FYL2X
end;


function Log2(X: Float): Float;
asm
   FLD1
   FLD      X
   FYL2X
end;

function LogN(Base, X: Float): Float;
asm
   FLD1
   FLD      X
   FYL2X
   FLD1
   FLD      Base
   FYL2X
   FDIV
end;

function IntPower(Base: Float; Exponent: Long): Float;
asm
   MOV      ECX, EAX
   CDQ
   FLD1
   XOR      EAX, EDX
   SUB      EAX, EDX
   JZ       @@3
   FLD      Base
   JMP      @@2
@@1:
   FMUL     ST, ST
@@2:
   SHR      EAX,1
   JNC      @@1
   FMUL     ST(1),ST
   JNZ      @@1
   FSTP     ST
   CMP      ECX, 0
   JGE      @@3
   FLD1
   FDIVRP
@@3:
   FWAIT
end;

function Power(Base, Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    if (Base = 0.0) and (Exponent > 0.0) then
      Result := 0.0
    else
      if (Frac(Exponent) = 0.0) and (Abs(Exponent) <= MaxInt) then
        Result := IntPower(Base, (Trunc(Exponent)))
      else
        Result := Exp(Exponent * Ln(Base))
end;

function Power10(Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    Result := Exp(Exponent * Ln10);
end;

function Power2(Exponent: Float): Float;
begin
  if Exponent = 0.0 then
    Result := 1.0
  else
    Result := Exp(Exponent * Ln2);
end;

function NearestPower2(Value:Long):Long;
begin
 {If you want to compute this for any value
 use this code. But it is 25x slower than
 this "case Value of" thing

 var S:Float
 S:=Log2(Value);

 if Trunc(S)<>S then
  S:=Trunc(S)+1;

 Result:=Round(Power2(S));}
 case Abs(Value) of
 0..2:Result:=2;
 3..4:Result:=4;
 5..8:Result:=8;
 9..16:Result:=16;
 17..32:Result:=32;
 33..64:Result:=64;
 65..128:Result:=128;
 129..256:Result:=256;
 257..512:Result:=512;
 513..1024:Result:=1024;
 1025..2048:Result:=2048;
 2049..4096:Result:=4096;
 4097..8192:Result:=8192;
 8193..16384:Result:=16384;
 16385..32768:Result:=32768;
 else
  Result:=65536;
 end;
end;


{*************************************************************************}
{                      Random number functions                            }
{*************************************************************************}


function RandRange(Min, Max: Long): Long;
begin
  if Min > Max then
    Result := Random(Min - Max) + Max
  else
    Result := Random(Max - Min) + Min;
end;

function RandRange(Min, Max: Float): Float;
begin
  if Min > Max then
    Result := Random(Round(Min - Max)) + Max
  else
    Result := Random(Round(Max - Min)) + Min;
end;

function RandSign(Value: Long): Long;
begin
  if Rand(2) = 0 then
    Result := Value
  else
    Result := -Value;
end;

function RandSign(Value: Float): Float;
begin
  if Rand(2) = 0 then
    Result := Value
  else
    Result := -Value;
end;

function RandomFromI(const Values: array of Long): Long;
begin
  Result := Values[Random(High(Values) + 1)];
end;

function RandomFromF(const Values: array of Float): Float;
begin
  Result := Values[Random(High(Values) + 1)];
end;


{*************************************************************************}
{                         Trigonometric functions                         }
{*************************************************************************}


function DegToRad(Degrees : Float) : Float;
begin
 Result:=Degrees * PIdiv180;
end;

function RadToDeg(Radians: Float): Float;
begin
 Result:=Radians * D180divPI;
end;

procedure SinCos(Angle: Float; var Sin, Cos: Double);
asm
   FLD      Angle
   FSINCOS
   FSTP     QWORD PTR [EDX]
   FSTP     QWORD PTR [EAX]
end;

procedure SinCos(Angle: Float; var Sin, Cos: Single);
asm
   FLD      Angle
   FSINCOS
   FSTP     DWORD PTR [EDX]
   FSTP     DWORD PTR [EAX]
end;

function Tan(Angle : Float) : Float;
asm
   FLD      Angle
   FPTAN
   FSTP     ST(0)
end;

function CoTan(Angle : Float) : Float;
asm
   FLD      Angle
   FPTAN
   FDIVRP
end;

function ArcSin(Value : Float) : Float;
asm
   FLD      Value
   FLD      ST
   FMUL     ST, ST
   FSUBR    One
   FSQRT
   FPATAN
end;

function ArcCos(Value : Float):Float;
asm
   FLD      Value
   FMUL     ST, ST
   FSUBR    One
   FSQRT
   FLD      Value
   FPATAN
end;

function ArcTan(Value: Float): Float;
asm
   FLD      Value
   FLD1
   FPATAN
   FWAIT
end;

function ArcCoTan(Value: Float): Float;
asm
   FLD1
   FLD      Value
   FPATAN
   FWAIT
end;

end.

